# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Wrappers, constants, and helpers around IOAdapter and its children."""
import pypowervm.const as pc
import pypowervm.util as u
import pypowervm.wrappers.entry_wrapper as ewrap

# Constants for generic I/O Adapter
IO_ADPT_ROOT = 'IOAdapter'
_IO_ADPT_ID = 'AdapterID'
_IO_ADPT_DESC = 'Description'
_IO_ADPT_DEV_NAME = 'DeviceName'
_IO_ADPT_DEV_TYPE = 'DeviceType'
_IO_ADPT_DYN_NAME = 'DynamicReconfigurationConnectorName'
_IO_ADPT_PHYS_LOC = 'PhysicalLocation'
_IO_ADPT_UDID = 'UniqueDeviceID'
_IO_ADPT_CHOICE = 'IOAdapterChoice'

# SR-IOV Adapter constants
_SRIOV_ADAPTER_ID = 'SRIOVAdapterID'
_SRIOV_ADAPTER_MODE = 'AdapterMode'
_SRIOV_ADAPTER_STATE = 'AdapterState'

_SRIOV_CONVERGED_ETHERNET_PHYSICAL_PORTS = 'ConvergedEthernetPhysicalPorts'
_SRIOV_ETHERNET_PHYSICAL_PORTS = 'EthernetPhysicalPorts'

# SR-IOV physical port constants

_SRIOVPP_CFG_SPEED = 'ConfiguredConnectionSpeed'
_SRIOVPP_CFG_FLOWCTL = 'ConfiguredFlowControl'
_SRIOVPP_CFG_MTU = 'ConfiguredMTU'
_SRIOVPP_CFG_OPTIONS = 'ConfiguredOptions'
_SRIOVPP_CFG_SWMODE = 'ConfiguredPortSwitchMode'
_SRIOVPP_CURR_SPEED = 'CurrentConnectionSpeed'
_SRIOVPP_CURR_OPTIONS = 'CurrentOptions'
_SRIOVPP_LBL = 'Label'
_SRIOVPP_LOC_CODE = 'LocationCode'
_SRIOVPP_MAX_DIAG_LPS = 'MaximumDiagnosticsLogicalPorts'
_SRIOVPP_MAX_PROM_LPS = 'MaximumPromiscuousLogicalPorts'
_SRIOVPP_ID = 'PhysicalPortID'
_SRIOVPP_CAPABILITIES = 'PortCapabilities'
_SRIOVPP_TYPE = 'PortType'
_SRIOVPP_LP_LIMIT = 'PortLogicalPortLimit'
_SRIOVPP_SUBLBL = 'SubLabel'
_SRIOVPP_SUPP_SPEEDS = 'SupportedConnectionSpeeds'
_SRIOVPP_SUPP_MTUS = 'SupportedMTUs'
_SRIOVPP_SUPP_OPTIONS = 'SupportedOptions'
_SRIOVPP_SUPP_PRI_ACL = 'SupportedPriorityAccessControlList'
_SRIOVPP_LINK_STATUS = 'LinkStatus'
_SRIOVPP_DEF_PORTVF_CFG = 'DefaultPortVFConfigurationOption'
_SRIOVPP_SEL_PORTVF_CFG = 'SelectedPortVFConfigurationOption'
_SRIOVPP_SUPP_PORTVF_CFG = 'SupportedPortVFConfigurationOptions'
_SRIOVPP_ALLOC_CAPACITY = 'AllocatedCapacity'
_SRIOVPP_CFG_MAX_ETHERNET_LPS = 'ConfiguredMaxEthernetLogicalPorts'
_SRIOVPP_CFG_ETHERNET_LPS = 'ConfiguredEthernetLogicalPorts'
_SRIOVPP_MAX_PVID = 'MaximumPortVLANID'
_SRIOVPP_MAX_VLAN_ID = 'MaximumVLANID'
_SRIOVPP_MIN_ETHERNET_CAPACITY_GRAN = 'MinimumEthernetCapacityGranularity'
_SRIOVPP_MIN_PVID = 'MinimumPortVLANID'
_SRIOVPP_MIN_VLAN_ID = 'MinimumVLANID'
_SRIOVPP_MAX_SUPP_ETHERNET_LPS = 'MaxSupportedEthernetLogicalPorts'
_SRIOVPP_MAX_ALLOW_ETH_VLANS = 'MaximumAllowedEthVLANs'
_SRIOVPP_MAX_ALLOW_ETH_MACS = 'MaximumAllowedEthMACs'
_SRIOVPP_SUPP_VLAN_RESTR = 'SupportedVLANRestrictions'
_SRIOVPP_SUPP_MAC_RESTR = 'SupportedMACRestrictions'
_SRIOVPP_CFG_MX_FCOE_LPS = 'ConfiguredMaxFiberChannelOverEthernetLogicalPorts'
_SRIOVPP_DEF_FCTARG_BACK_DEV = 'DefaultFiberChannelTargetsForBackingDevice'
_SRIOVPP_DEF_FTARG_NBACK_DEV = 'DefaultFiberChannelTargetsForNonBackingDevice'
_SRIOVPP_CFG_FCOE_LPS = 'ConfiguredFiberChannelOverEthernetLogicalPorts'
_SRIOVPP_MIN_FCOE_CAPACITY_GRAN = 'MinimumFCoECapacityGranularity'
_SRIOVPP_FC_TARGET_ROUNDING_VALUE = 'FiberChannelTargetsRoundingValue'
_SRIOVPP_MX_SUPP_FCOE_LPS = 'MaxSupportedFiberChannelOverEthernetLogicalPorts'
_SRIOVPP_MAX_FC_TARGETS = 'MaximumFiberChannelTargets'

_SRIOVPP_EL_ORDER = (
    _SRIOVPP_CFG_SPEED, _SRIOVPP_CFG_FLOWCTL, _SRIOVPP_CFG_MTU,
    _SRIOVPP_CFG_OPTIONS, _SRIOVPP_CFG_SWMODE, _SRIOVPP_CURR_SPEED,
    _SRIOVPP_CURR_OPTIONS, _SRIOVPP_LBL, _SRIOVPP_LOC_CODE,
    _SRIOVPP_MAX_DIAG_LPS, _SRIOVPP_MAX_PROM_LPS,
    _SRIOVPP_ID, _SRIOVPP_CAPABILITIES, _SRIOVPP_TYPE,
    _SRIOVPP_LP_LIMIT, _SRIOVPP_SUBLBL, _SRIOVPP_SUPP_SPEEDS,
    _SRIOVPP_SUPP_MTUS, _SRIOVPP_SUPP_OPTIONS,
    _SRIOVPP_SUPP_PRI_ACL, _SRIOVPP_LINK_STATUS, _SRIOVPP_DEF_PORTVF_CFG,
    _SRIOVPP_SEL_PORTVF_CFG, _SRIOVPP_SUPP_PORTVF_CFG)

_SRIOVEPP_EL_ORDER = _SRIOVPP_EL_ORDER + (
    _SRIOVPP_ALLOC_CAPACITY,
    _SRIOVPP_CFG_MAX_ETHERNET_LPS,
    _SRIOVPP_CFG_ETHERNET_LPS, _SRIOVPP_MAX_PVID,
    _SRIOVPP_MAX_VLAN_ID, _SRIOVPP_MIN_ETHERNET_CAPACITY_GRAN,
    _SRIOVPP_MIN_PVID, _SRIOVPP_MIN_VLAN_ID,
    _SRIOVPP_MAX_SUPP_ETHERNET_LPS, _SRIOVPP_MAX_ALLOW_ETH_VLANS,
    _SRIOVPP_MAX_ALLOW_ETH_MACS, _SRIOVPP_SUPP_VLAN_RESTR,
    _SRIOVPP_SUPP_MAC_RESTR)

_SRIOVCPP_EL_ORDER = _SRIOVEPP_EL_ORDER + (
    _SRIOVPP_CFG_MX_FCOE_LPS,
    _SRIOVPP_DEF_FCTARG_BACK_DEV, _SRIOVPP_DEF_FTARG_NBACK_DEV,
    _SRIOVPP_CFG_FCOE_LPS, _SRIOVPP_MIN_FCOE_CAPACITY_GRAN,
    _SRIOVPP_FC_TARGET_ROUNDING_VALUE, _SRIOVPP_MX_SUPP_FCOE_LPS,
    _SRIOVPP_MAX_FC_TARGETS)

# SR-IOV logical port constants
_SRIOVLP_CFG_ID = 'ConfigurationID'
_SRIOVLP_ID = 'LogicalPortID'
_SRIOVLP_ADPT_ID = 'AdapterID'
_SRIOVLP_DRC_NAME = 'DynamicReconfigurationConnectorName'
_SRIOVLP_IS_FUNC = 'IsFunctional'
_SRIOVLP_IS_PROMISC = 'IsPromiscous'  # [sic]
_SRIOVLP_IS_DIAG = 'IsDiagnostic'
_SRIOVLP_IS_DEBUG = 'IsDebug'
_SRIOVLP_IS_HUGE_DMA = 'IsHugeDMA'
_SRIOVLP_DEV_NAME = 'DeviceName'
_SRIOVLP_CFG_CAPACITY = 'ConfiguredCapacity'
_SRIOVLP_PPORT_ID = 'PhysicalPortID'
_SRIOVLP_PVID = 'PortVLANID'
_SRIOVLP_LOC_CODE = 'LocationCode'
_SRIOVLP_TUNING_BUF_ID = 'TuningBufferID'
_SRIOVLP_VNIC_PORT_USAGE = 'VNICPortUsage'
_SRIOVLP_ASSOC_LPARS = 'AssociatedLogicalPartitions'
_SRIOVLP_ALLOWED_MACS = 'AllowedMACAddresses'
_SRIOVLP_MAC = 'MACAddress'
_SRIOVLP_CUR_MAC = 'CurrentMACAddress'
_SRIOVLP_8021Q_ALLOW_PRI = 'IEEE8021QAllowablePriorities'
_SRIOVLP_8021Q_PRI = 'IEEE8021QPriority'
_SRIOVLP_MAC_FLAGS = 'MACAddressFlags'
_SRIOVLP_NUM_ALLOWED_VLANS = 'NumberOfAllowedVLANs'
_SRIOVLP_ALLOWED_VLANS = 'AllowedVLANs'

_SRIOVLP_EL_ORDER = (
    _SRIOVLP_CFG_ID, _SRIOVLP_ID, _SRIOVLP_ADPT_ID, _SRIOVLP_DRC_NAME,
    _SRIOVLP_IS_FUNC, _SRIOVLP_IS_PROMISC, _SRIOVLP_IS_DIAG, _SRIOVLP_IS_DEBUG,
    _SRIOVLP_IS_HUGE_DMA, _SRIOVLP_DEV_NAME, _SRIOVLP_CFG_CAPACITY,
    _SRIOVLP_PPORT_ID, _SRIOVLP_PVID, _SRIOVLP_LOC_CODE,
    _SRIOVLP_TUNING_BUF_ID, _SRIOVLP_VNIC_PORT_USAGE, _SRIOVLP_ASSOC_LPARS,
    _SRIOVLP_ALLOWED_MACS, _SRIOVLP_MAC, _SRIOVLP_CUR_MAC,
    _SRIOVLP_8021Q_ALLOW_PRI, _SRIOVLP_8021Q_PRI, _SRIOVLP_MAC_FLAGS,
    _SRIOVLP_NUM_ALLOWED_VLANS, _SRIOVLP_ALLOWED_VLANS)

# Top-level VNIC properties
_VNIC_DED = 'VirtualNICDedicated'
_VNIC_ADP_TYPE = 'AdapterType'
_VNIC_DRC_NAME = 'DynamicReconfigurationConnectorName'
_VNIC_LOC_CODE = 'LocationCode'
_VNIC_LPAR_ID = 'LocalPartitionID'
_VNIC_REQ_ADP = 'RequiredAdapter'
_VNIC_VARIED_ON = 'VariedOn'
_VNIC_USE_NEXT_AVAIL_SLOT = 'UseNextAvailableSlotID'
_VNIC_USE_NEXT_AVAIL_HIGH_SLOT = 'UseNextAvailableHighSlotID'
_VNIC_SLOT_NUM = 'VirtualSlotNumber'
_VNIC_ENABLED = 'Enabled'
_VNIC_DETAILS = 'Details'
_VNIC_BACK_DEVS = 'AssociatedBackingDevices'

_VNIC_EL_ORDER = (
    _VNIC_DED, _VNIC_ADP_TYPE, _VNIC_DRC_NAME, _VNIC_LOC_CODE, _VNIC_LPAR_ID,
    _VNIC_REQ_ADP, _VNIC_VARIED_ON, _VNIC_USE_NEXT_AVAIL_SLOT,
    _VNIC_USE_NEXT_AVAIL_HIGH_SLOT, _VNIC_SLOT_NUM, _VNIC_ENABLED,
    _VNIC_DETAILS, _VNIC_BACK_DEVS)

# Properties for _VNICDetails (schema: VirtualNICDetails.Type)
_VNICD_PVID = 'PortVLANID'
_VNICD_PVID_PRI = 'PortVLANIDPriority'
_VNICD_ALLOWED_VLANS = 'AllowedVLANIDs'
_VNICD_MAC = 'MACAddress'
_VNICD_ALLOWED_OS_MACS = 'AllowedOperatingSystemMACAddresses'
_VNICD_OS_DEV_NAME = 'OSDeviceName'
_VNICD_DES_MODE = 'DesiredMode'
_VNICD_DES_CAP_PCT = 'DesiredCapacityPercentage'
_VNICD_AUTO_FB = 'AutoFailBack'

_VNICD_EL_ORDER = (
    _VNICD_PVID, _VNICD_PVID_PRI, _VNICD_ALLOWED_VLANS, _VNICD_MAC,
    _VNICD_ALLOWED_OS_MACS, _VNICD_OS_DEV_NAME, _VNICD_DES_MODE,
    _VNICD_DES_CAP_PCT, _VNICD_DES_CAP_PCT)

# Properties for VNICBackDev (schema: VirtualNICSRIOVBackingDevice)
_VNICBD_CHOICE = 'VirtualNICBackingDeviceChoice'
_VNICBD = 'VirtualNICSRIOVBackingDevice'
_VNICBD_DEV_TYP = 'DeviceType'
_VNICBD_VIOS = 'AssociatedVirtualIOServer'
_VNICBD_SWITCH = 'AssociatedVirtualNICSwitch'
_VNICBD_VNIC = 'AssociatedVirtualNICDedicated'
_VNICBD_ACTIVE = 'IsActive'
_VNICBD_STATUS = 'Status'
_VNICBD_FAILOVER_PRI = 'FailOverPriority'
_VNICBD_ACTION = 'BackingDeviceAction'
_VNICBD_SRIOV_ADP_ID = 'RelatedSRIOVAdapterID'
_VNICBD_CUR_CAP_PCT = 'CurrentCapacityPercentage'
_VNICBD_PPORT_ID = 'RelatedSRIOVPhysicalPortID'
_VNICBD_LPORT = 'RelatedSRIOVLogicalPort'
_VNICBD_DES_CAP_PCT = 'DesiredCapacityPercentage'
# For building the VIOS HREF.  (Would have liked to use pypowervm.wrappers.
# virtual_io_server.VIOS.schema_type, but circular import.)
_VIOS = 'VirtualIOServer'

_VNICBD_EL_ORDER = (
    _VNICBD_DEV_TYP, _VNICBD_VIOS, _VNICBD_SWITCH, _VNICBD_VNIC,
    _VNICBD_ACTIVE, _VNICBD_STATUS, _VNICBD_FAILOVER_PRI, _VNICBD_ACTION,
    _VNICBD_SRIOV_ADP_ID, _VNICBD_CUR_CAP_PCT, _VNICBD_PPORT_ID, _VNICBD_LPORT,
    _VNICBD_DES_CAP_PCT)

# Physical Fibre Channel Port Constants
_PFC_PORT_LOC_CODE = 'LocationCode'
_PFC_PORT_NAME = 'PortName'
_PFC_PORT_UDID = 'UniqueDeviceID'
PFC_PORT_WWPN = 'WWPN'
_PFC_PORT_AVAILABLE_PORTS = 'AvailablePorts'
_PFC_PORT_TOTAL_PORTS = 'TotalPorts'
PFC_PORTS_ROOT = 'PhysicalFibreChannelPorts'
PFC_PORT_ROOT = 'PhysicalFibreChannelPort'


class SRIOVAdapterMode(object):
    """Enumeration for SR-IOV adapter modes (from SRIOVAdapterMode.Enum)."""
    SRIOV = 'Sriov'
    DEDICATED = 'Dedicated'
    FORCE_DEDICATED = 'ForceDedicated'
    UNKNOWN = 'unknown'


class SRIOVAdapterState(object):
    """Enumeration for SR-IOV adapter states (from SRIOVAdapterState.Enum)."""
    INITIALIZING = 'Initializing'
    NOT_CONFIG = 'NotConfigured'
    POWERED_OFF = 'PoweredOff'
    POWERING_OFF = 'PoweringOff'
    RUNNING = 'Running'
    DUMPING = 'Dumping'
    FAILED = 'Failed'
    MISSING = 'Missing'
    MISMATCH = 'PCIEIDMismatch'


class SRIOVSpeed(object):
    """Enumeration for SR-IOV speed (from SRIOVConnectionSpeed.Enum)."""
    E10M = 'E10Mbps'
    E100M = 'E100Mbps'
    E1G = 'E1Gbps'
    E10G = 'E10Gbps'
    E40G = 'E40Gbps'
    E100G = 'E100Gbps'
    AUTO = 'Auto'
    UNKNOWN = 'Unknown'


class SRIOVPPMTU(object):
    """SR-IOV Phys Port Max Transmission Unit (SRIOVPhysicalPortMTU.Enum)."""
    E1500 = "E_1500"
    E9000 = "E_9000"
    UNKNOWN = 'Unknown'


class VNICBackDevStatus(object):
    """Enumeration of possible VNIC backing device statuses."""
    OPERATIONAL = 'OPERATIONAL'
    POWERED_OFF = 'POWERED_OFF'
    LINK_DOWN = 'LINK_DOWN'
    NETWORK_ERROR = 'NETWORK_ERROR'
    UNRESPONSIVE = 'UNRESPONSIVE'
    ADAPTER_ERROR = 'ADAPTER_ERROR'
    UNKNOWN = 'UNKNOWN'


class VNICPortUsage(object):
    """Enumeration of possible VNIC port usages."""
    NOT_VNIC = 'NOT_VNIC'
    DEDICATED_VNIC = 'DEDICATED_VNIC'
    SHARED_VNIC = 'SHARED_VNIC'


@ewrap.ElementWrapper.pvm_type(IO_ADPT_ROOT, has_metadata=True)
class IOAdapter(ewrap.ElementWrapper):
    """A generic IO Adapter.

    This is a device plugged in to the system.  The location code indicates
    where it is plugged into the system.
    """

    @property
    def id(self):
        """The adapter system id."""
        return self._get_val_str(_IO_ADPT_ID)

    @property
    def description(self):
        return self._get_val_str(_IO_ADPT_DESC)

    @property
    def dev_name(self):
        return self._get_val_str(_IO_ADPT_DEV_NAME)

    @property
    def dev_type(self):
        return self._get_val_str(_IO_ADPT_DEV_TYPE)

    @property
    def drc_name(self):
        return self._get_val_str(_IO_ADPT_DYN_NAME)

    @property
    def phys_loc_code(self):
        return self._get_val_str(_IO_ADPT_PHYS_LOC)

    @property
    def udid(self):
        return self._get_val_str(_IO_ADPT_UDID)


@ewrap.ElementWrapper.pvm_type('PhysicalFibreChannelAdapter',
                               has_metadata=True)
class PhysFCAdapter(IOAdapter):
    """A Physical Fibre Channel I/O Adapter.

    Extends the generic I/O Adapter, but provides port detail as well.
    The adapter has a set of Physical Fibre Channel Ports (PhysFCPort).
    """

    @property
    def fc_ports(self):
        """A set of Physical Fibre Channel Ports.

        The set of PhysFCPort's that are attached to this adapter.
        The data on this should be considered read only.
        """
        es = ewrap.WrapperElemList(self._find_or_seed(PFC_PORTS_ROOT),
                                   PhysFCPort)
        return es


@ewrap.ElementWrapper.pvm_type('PhysicalFibreChannelPort', has_metadata=True)
class PhysFCPort(ewrap.ElementWrapper):
    """A Physical Fibre Channel Port."""

    @classmethod
    def bld_ref(cls, adapter, name, ref_tag=None):
        """Create a wrapper that serves as a reference to a port.

        This is typically used when another element (ex. Virtual FC Mapping)
        requires a port to be specified in it.  Rather than query to find
        the port, one can simply be built and referenced as a child element.
        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param name: The name of the physical FC port.  End users need to
                     verify the port name.  Typically starts with 'fcs'.
        :param ref_tag: (Optional, Default=None) If set, override the
                        default 'PhysicalFibreChannelPort' tag/label in the
                        element with the string specified.
        """
        port = super(PhysFCPort, cls)._bld(adapter)
        port._name(name)
        if ref_tag:
            port.element.tag = ref_tag
        return port

    @property
    def loc_code(self):
        return self._get_val_str(_PFC_PORT_LOC_CODE)

    @property
    def name(self):
        return self._get_val_str(_PFC_PORT_NAME)

    def _name(self, value):
        return self.set_parm_value(_PFC_PORT_NAME, value)

    @property
    def udid(self):
        return self._get_val_str(_PFC_PORT_UDID)

    @property
    def wwpn(self):
        return self._get_val_str(PFC_PORT_WWPN)

    @property
    def npiv_available_ports(self):
        return self._get_val_int(_PFC_PORT_AVAILABLE_PORTS, 0)

    @property
    def npiv_total_ports(self):
        return self._get_val_int(_PFC_PORT_TOTAL_PORTS, 0)


@ewrap.ElementWrapper.pvm_type('SRIOVAdapter', has_metadata=True)
class SRIOVAdapter(IOAdapter):
    """The SR-IOV adapters for this system."""

    @property
    def sriov_adap_id(self):
        """Not to be confused with the 'id' property (IOAdapter.AdapterID)."""
        return self._get_val_int(_SRIOV_ADAPTER_ID)

    @property
    def mode(self):
        return self._get_val_str(_SRIOV_ADAPTER_MODE)

    @mode.setter
    def mode(self, value):
        self.set_parm_value(_SRIOV_ADAPTER_MODE, value)

    @property
    def state(self):
        return self._get_val_str(_SRIOV_ADAPTER_STATE)

    def _convergedphysicalports(self):
        """Retrieve all Converged physical ports."""
        elem = self._find(_SRIOV_CONVERGED_ETHERNET_PHYSICAL_PORTS)
        if elem is None:
            return None
        return ewrap.WrapperElemList(elem, child_class=SRIOVConvPPort)

    def _ethernetphysicalports(self):
        """Retrieve all Ethernet physical ports."""
        elem = self._find(_SRIOV_ETHERNET_PHYSICAL_PORTS)
        if elem is None:
            return None
        return ewrap.WrapperElemList(elem, child_class=SRIOVEthPPort)

    @property
    def phys_ports(self):
        """Retrieve Combined list of all physical ports.

        Returns a list of converged and ethernet physical ports.
        This list is not modifiable, cannot insert or remove
        items from it, however, individual item can be updated.
        For example, label and sublabels can be updated.
        """
        allports = []
        cports = self._convergedphysicalports()
        eports = self._ethernetphysicalports()
        for c in cports or []:
            allports.append(c)
        for e in eports or []:
            allports.append(e)
        # Set the ports' backpointers to this SRIOVAdapter
        for pport in allports:
            pport._sriov_adap = self
        return allports


@ewrap.ElementWrapper.pvm_type('SRIOVEthernetPhysicalPort', has_metadata=True,
                               child_order=_SRIOVEPP_EL_ORDER)
class SRIOVEthPPort(ewrap.ElementWrapper):
    """The SR-IOV Ethernet Physical port."""

    def __init__(self):
        super(SRIOVEthPPort, self).__init__()
        # This must be set by the instantiating SRIOVAdapter.
        self._sriov_adap = None

    @property
    def sriov_adap(self):
        """Backpointer to the SRIOVAdapter owning this physical port."""
        if self._sriov_adap is None:
            raise NotImplementedError("Developer error: SRIOVAdapter pointer "
                                      "not set!")
        return self._sriov_adap

    @property
    def sriov_adap_id(self):
        """The integer sriov_adap_id of the SRIOVAdapter owning this port."""
        return self.sriov_adap.sriov_adap_id

    @property
    def label(self):
        return self._get_val_str(_SRIOVPP_LBL)

    @label.setter
    def label(self, value):
        self.set_parm_value(_SRIOVPP_LBL, value)

    @property
    def loc_code(self):
        return self._get_val_str(_SRIOVPP_LOC_CODE)

    @property
    def port_id(self):
        return self._get_val_int(_SRIOVPP_ID)

    @property
    def sublabel(self):
        return self._get_val_str(_SRIOVPP_SUBLBL)

    @sublabel.setter
    def sublabel(self, value):
        self.set_parm_value(_SRIOVPP_SUBLBL, value)

    @property
    def link_status(self):
        return self._get_val_bool(_SRIOVPP_LINK_STATUS)

    @property
    def cfg_max_lps(self):
        return self._get_val_int(_SRIOVPP_CFG_MAX_ETHERNET_LPS)

    @cfg_max_lps.setter
    def cfg_max_lps(self, value):
        self.set_parm_value(_SRIOVPP_CFG_MAX_ETHERNET_LPS, value)

    @property
    def cfg_lps(self):
        return self._get_val_int(_SRIOVPP_CFG_ETHERNET_LPS)

    @property
    def min_granularity(self):
        """Gets the minimum granularity in a float-percentage format.

        :return: If the property is say "2.45%", a value of .0245 will be
                 returned.
        """
        return self._get_val_percent(_SRIOVPP_MIN_ETHERNET_CAPACITY_GRAN)

    @property
    def supp_max_lps(self):
        return self._get_val_int(_SRIOVPP_MAX_SUPP_ETHERNET_LPS)

    @property
    def allocated_capacity(self):
        """Gets the allocated capacity in a float-percentage format.

        :return: If the property is say "2.45%", a value of .0245 will be
                 returned.
        """
        return self._get_val_percent(_SRIOVPP_ALLOC_CAPACITY)

    @property
    def curr_speed(self):
        return self._get_val_str(_SRIOVPP_CURR_SPEED)

    @property
    def mtu(self):
        """Result should be a SRIOVPPMTU value."""
        return self._get_val_str(_SRIOVPP_CFG_MTU)

    @mtu.setter
    def mtu(self, val):
        """Input val should be a SRIOVPPMTU value."""
        self.set_parm_value(_SRIOVPP_CFG_MTU, val)

    @property
    def switch_mode(self):
        """Result should be a network.VSwitchMode value."""
        return self._get_val_str(_SRIOVPP_CFG_SWMODE)

    @switch_mode.setter
    def switch_mode(self, val):
        """Input val should be a network.VSwitchMode value."""
        self.set_parm_value(_SRIOVPP_CFG_SWMODE, val)

    @property
    def flow_ctl(self):
        return self._get_val_bool(_SRIOVPP_CFG_FLOWCTL)

    @flow_ctl.setter
    def flow_ctl(self, val):
        self.set_parm_value(_SRIOVPP_CFG_FLOWCTL, u.sanitize_bool_for_api(val))


@ewrap.ElementWrapper.pvm_type('SRIOVConvergedNetworkAdapterPhysicalPort',
                               has_metadata=True,
                               child_order=_SRIOVCPP_EL_ORDER)
class SRIOVConvPPort(SRIOVEthPPort):
    """The SR-IOV Converged Physical port."""
    pass


@ewrap.EntryWrapper.pvm_type('SRIOVEthernetLogicalPort',
                             child_order=_SRIOVLP_EL_ORDER)
class SRIOVEthLPort(ewrap.EntryWrapper):
    """The SR-IOV Ethernet Logical port."""

    @classmethod
    def bld(cls, adapter, sriov_adap_id, pport_id, pvid=None, mac=None,
            allowed_vlans=u.VLANList.ALL, allowed_macs=u.MACList.ALL,
            is_promisc=False, cfg_capacity=None):
        """Create a wrapper used to create a logical port on the server.

        :param adapter: A pypowervm.adapter.Adapter (for traits, etc.)
        :param sriov_adap_id: Corresponds to SRIOVAdapter.SRIOVAdapterID,
                              *not* SRIOVAdapter.AdapterID
        :param pport_id: The physical port ID this logical port is part of.
        :param pvid: The port VLAN identifier for this logical port. Any
                     untagged traffic passing through this port will have
                     this VLAN tag added.
        :param mac: The MAC address to assign to the logical port.
        :param allowed_vlans: An integer list of VLANS allowed on this logical
                              port. Specify pypowervm.util.VLANList.ALL to
                              allow all VLANs or .NONE to allow no VLANs on
                              this logical port. Default: ALL.
        :param allowed_macs: List of string MAC addresses allowed on this
                             logical port. Specify pypowervm.util.MACList.ALL
                             to allow all MAC addresses, or .NONE to allow no
                             MAC addresses on this logical port.  Default: ALL.
        :param is_promisc: Set to True if using the logical port for bridging
                           (e.g. SEA, OVS, etc.); False if assigning directly
                           to an LPAR.  Only one logical port per physical port
                           may be promiscuous.
        :param cfg_capacity: The configured capacity of the logical port as a
                             percentage.  This represents the minimum bandwidth
                             this logical port will receive, as a percentage
                             of bandwidth available from the physical port.
                             The valid values are 0.0 <= x <= 1.0 up to 2
                             decimal places.  This will be intrepreted as a
                             percentage, where 0.02 == 2%.
        """
        lport = super(SRIOVEthLPort, cls)._bld(adapter)
        lport._sriov_adap_id(sriov_adap_id)
        lport._pport_id(pport_id)
        if pvid is not None:
            lport.pvid = pvid
        lport.allowed_vlans = allowed_vlans
        if mac is not None:
            lport._mac(mac)
        lport.allowed_macs = allowed_macs
        lport._is_promisc(is_promisc)
        if cfg_capacity:
            lport._cfg_capacity(cfg_capacity)
        return lport

    @property
    def lport_id(self):
        return self._get_val_int(_SRIOVLP_ID)

    @property
    def sriov_adap_id(self):
        return self._get_val_int(_SRIOVLP_ADPT_ID)

    def _sriov_adap_id(self, value):
        self.set_parm_value(_SRIOVLP_ADPT_ID, value)

    @property
    def is_promisc(self):
        return self._get_val_bool(_SRIOVLP_IS_PROMISC)

    def _is_promisc(self, value):
        self.set_parm_value(_SRIOVLP_IS_PROMISC,
                            u.sanitize_bool_for_api(value))

    @property
    def dev_name(self):
        return self._get_val_str(_SRIOVLP_DEV_NAME)

    @property
    def cfg_capacity(self):
        """Gets the configured capacity in a float-percentage format.

        :return: If the property is say "2.45%", a value of .0245 will be
                 returned.
        """
        return self._get_val_percent(_SRIOVLP_CFG_CAPACITY)

    def _cfg_capacity(self, value):
        """The configured capacity

        :param value: The configured capacity value.  The valid values are
                      0.0 <= x <=1.0 up to 2 decimal places.  This will be
                      intrepreted as a percentage, where 0.02 == 2%.
        """
        self.set_parm_value(_SRIOVLP_CFG_CAPACITY,
                            u.sanitize_percent_for_api(value))

    @property
    def pport_id(self):
        """Gets the physical port short ID."""
        return self._get_val_int(_SRIOVLP_PPORT_ID)

    def _pport_id(self, value):
        """Internal setter for the physical port short ID."""
        self.set_parm_value(_SRIOVLP_PPORT_ID, value)

    @property
    def pvid(self):
        return self._get_val_int(_SRIOVLP_PVID)

    @pvid.setter
    def pvid(self, value):
        self.set_parm_value(_SRIOVLP_PVID, value)

    @property
    def allowed_vlans(self):
        vlan_str = self._get_val_str(_SRIOVLP_ALLOWED_VLANS)
        return u.VLANList.unmarshal(vlan_str) if vlan_str is not None else None

    @allowed_vlans.setter
    def allowed_vlans(self, vlans):
        self.set_parm_value(_SRIOVLP_ALLOWED_VLANS, u.VLANList.marshal(vlans))

    @property
    def mac(self):
        """MAC address of the format XXXXXXXXXXXX (12 uppercase hex digits).

        This is the MAC address "burned into" the logical port.  The actual MAC
        address on the interface (cur_mac) may be this value or the value set
        from within the OS on the VM.
        """
        return self._get_val_str(_SRIOVLP_MAC)

    def _mac(self, value):
        self.set_parm_value(_SRIOVLP_MAC, u.sanitize_mac_for_api(value))

    @property
    def allowed_macs(self):
        amstr = self._get_val_str(_SRIOVLP_ALLOWED_MACS)
        return u.MACList.unmarshal(amstr) if amstr is not None else None

    @allowed_macs.setter
    def allowed_macs(self, maclist):
        self.set_parm_value(_SRIOVLP_ALLOWED_MACS, u.MACList.marshal(maclist))

    @property
    def cur_mac(self):
        """MAC address of the format XXXXXXXXXXXX (12 uppercase hex digits).

        This is the real value set on the interface, possibly by the VM's OS.

        Note that some SR-IOV cards are broken and don't report the OS-assigned
        value correctly.  In such cases, cur_mac will report the same as mac.
        """
        return self._get_val_str(_SRIOVLP_CUR_MAC)

    @property
    def loc_code(self):
        return self._get_val_str(_SRIOVLP_LOC_CODE)

    @property
    def vnic_port_usage(self):
        return self._get_val_str(_SRIOVLP_VNIC_PORT_USAGE)


@ewrap.EntryWrapper.pvm_type(_VNIC_DED, child_order=_VNIC_EL_ORDER)
class VNIC(ewrap.EntryWrapper):
    """A dedicated, possibly-redundant Virtual NIC."""

    @classmethod
    def bld(cls, adapter, pvid=None, slot_num=None,
            allowed_vlans=u.VLANList.ALL, mac_addr=None,
            allowed_macs=u.MACList.ALL, back_devs=None):
        """Build a new VNIC wrapper suitable for .create()

        A VNIC is a CHILD object on a LogicalPartition.  Usage models:
            vnic = VNIC.bld(...)
            vnic.back_devs.append(back_dev1)
            ...
        or
            vnic = VNIC.bld(..., back_devs=[back_dev1, back_dev2, ...])
        then
            vnic.create(parent=lpar_wrap)

        :param adapter: pypowervm.adapter.Adapter for REST API communication.
        :param pvid: Port VLAN ID for this vNIC.  If not specified, the vNIC's
                     traffic is untagged.
        :param slot_num: Desired virtual slot number on the owning LPAR.
        :param allowed_vlans: An integer list of VLANS allowed on this vNIC.
                              Specify pypowervm.util.VLANList.ALL to allow all
                              VLANs or .NONE to allow no VLANs on this vNIC.
                              Default: ALL.
        :param mac_addr: MAC address for the vNIC.
        :param allowed_macs: List of string MAC addresses allowed on this vNIC.
                             Specify pypowervm.util.MACList.ALL to allow all
                             MAC addresses, or .NONE to allow no MAC addresses
                             on this vNIC.  Default: ALL.
        :param back_devs: List of VNICBackDev wrappers each indicating a
                          combination of VIOS, SR-IOV adapter and physical port
                          on which to create the VF for the backing device.
                          See VNICBackDev.bld.  If not specified to bld, at
                          least one must be added before the VNIC can be
                          created.
        :return: A new VNIC wrapper.
        """
        vnic = super(VNIC, cls)._bld(adapter)
        if slot_num is not None:
            vnic._slot(slot_num)
        else:
            vnic._use_next_avail_slot_id = True

        vnic._details = _VNICDetails._bld_new(
            adapter, pvid=pvid, allowed_vlans=allowed_vlans, mac_addr=mac_addr,
            allowed_macs=allowed_macs)

        if back_devs:
            vnic.back_devs = back_devs
        return vnic

    @property
    def drc_name(self):
        return self._get_val_str(_VNIC_DRC_NAME)

    @property
    def lpar_id(self):
        """The integer ID, not UUID, of the LPAR owning this VNIC."""
        return self._get_val_int(_VNIC_LPAR_ID)

    @property
    def slot(self):
        return self._get_val_int(_VNIC_SLOT_NUM)

    def _slot(self, val):
        self.set_parm_value(_VNIC_SLOT_NUM, val)

    @property
    def _use_next_avail_slot_id(self):
        """Use next available (high) slot ID, true or false."""
        unasi_field = (_VNIC_USE_NEXT_AVAIL_HIGH_SLOT
                       if self.traits.has_high_slot
                       else _VNIC_USE_NEXT_AVAIL_SLOT)
        return self._get_val_bool(unasi_field)

    @_use_next_avail_slot_id.setter
    def _use_next_avail_slot_id(self, unasi):
        """Use next available (high) slot ID.

        :param unasi: Boolean value to set (True or False)
        """
        unasi_field = (_VNIC_USE_NEXT_AVAIL_HIGH_SLOT
                       if self.traits.has_high_slot
                       else _VNIC_USE_NEXT_AVAIL_SLOT)
        self.set_parm_value(unasi_field, u.sanitize_bool_for_api(unasi))

    @property
    def pvid(self):
        """The integer port VLAN ID, or None if the vNIC has no PVID."""
        return self._details.pvid

    @pvid.setter
    def pvid(self, val):
        self._details.pvid = val

    @property
    def allowed_vlans(self):
        return self._details.allowed_vlans

    @allowed_vlans.setter
    def allowed_vlans(self, vlans):
        self._details.allowed_vlans = vlans

    @property
    def mac(self):
        """MAC address of the format XXXXXXXXXXXX (12 uppercase hex digits)."""
        return self._details.mac

    def _mac(self, val):
        self._details._mac(val)

    @property
    def allowed_macs(self):
        return self._details.allowed_macs

    @allowed_macs.setter
    def allowed_macs(self, maclist):
        self._details.allowed_macs = maclist

    @property
    def capacity(self):
        """The capacity (float, 0.0-1.0) of the active backing logical port."""
        return self._details.capacity

    @property
    def _details(self):
        return _VNICDetails.wrap(self._find_or_seed(_VNIC_DETAILS))

    @_details.setter
    def _details(self, val):
        self.element.replace(self._find_or_seed(_VNIC_DETAILS), val.element)

    @property
    def back_devs(self):
        return ewrap.WrapperElemList(self._find_or_seed(_VNIC_BACK_DEVS),
                                     child_class=VNICBackDev,
                                     indirect=_VNICBD_CHOICE)

    @back_devs.setter
    def back_devs(self, new_devs):
        self.replace_list(_VNIC_BACK_DEVS, new_devs, indirect=_VNICBD_CHOICE)

    @property
    def auto_pri_failover(self):
        return self._details.auto_pri_failover

    @auto_pri_failover.setter
    def auto_pri_failover(self, val):
        self._details.auto_pri_failover = val


@ewrap.ElementWrapper.pvm_type(_VNIC_DETAILS, has_metadata=True,
                               child_order=_VNICD_EL_ORDER)
class _VNICDetails(ewrap.ElementWrapper):
    """The 'Details' sub-element of a VirtualNICDedicated."""

    @classmethod
    def _bld_new(cls, adapter, pvid=None, allowed_vlans=u.VLANList.ALL,
                 mac_addr=None, allowed_macs=u.MACList.ALL):
        """Create a new _VNICDetails wrapper suitable for insertion into a VNIC.

        Not to be called outside of VNIC.bld().

        :param adapter: pypowervm.adapter.Adapter for REST API communication.
        :param pvid: Port VLAN ID for this vNIC.  If not specified, the vNIC's
                     traffic is untagged.
        :param allowed_vlans: An integer list of VLANS allowed on this vNIC.
                              Specify pypowervm.util.VLANList.ALL to allow all
                              VLANs or .NONE to allow no VLANs on this vNIC.
                              Default: ALL.
        :param mac_addr: MAC address for the vNIC.
        :param allowed_macs: List of string MAC addresses allowed on this vNIC.
                             Specify pypowervm.util.MACList.ALL to allow all
                             MAC addresses, or .NONE to allow no MAC addresses
                             on this vNIC.  Default: ALL.
        :return: A new _VNICDetails wrapper.
        """
        vnicd = super(_VNICDetails, cls)._bld(adapter)
        if pvid is not None:
            vnicd.pvid = pvid
        vnicd.allowed_vlans = allowed_vlans
        if mac_addr is not None:
            vnicd._mac(mac_addr)
        vnicd.allowed_macs = allowed_macs
        return vnicd

    @property
    def pvid(self):
        """The integer port VLAN ID, or None if the vNIC has no PVID."""
        return self._get_val_int(_VNICD_PVID)

    @pvid.setter
    def pvid(self, val):
        self.set_parm_value(_VNICD_PVID, val)

    @property
    def allowed_vlans(self):
        vlan_str = self._get_val_str(_VNICD_ALLOWED_VLANS)
        return u.VLANList.unmarshal(vlan_str) if vlan_str is not None else None

    @allowed_vlans.setter
    def allowed_vlans(self, vlans):
        self.set_parm_value(_VNICD_ALLOWED_VLANS, u.VLANList.marshal(vlans))

    @property
    def mac(self):
        """MAC address of the format XXXXXXXXXXXX (12 uppercase hex digits)."""
        return self._get_val_str(_VNICD_MAC)

    def _mac(self, val):
        self.set_parm_value(_VNICD_MAC, u.sanitize_mac_for_api(val))

    @property
    def allowed_macs(self):
        amstr = self._get_val_str(_VNICD_ALLOWED_OS_MACS)
        return u.MACList.unmarshal(amstr) if amstr is not None else None

    @allowed_macs.setter
    def allowed_macs(self, maclist):
        self.set_parm_value(_VNICD_ALLOWED_OS_MACS, u.MACList.marshal(maclist))

    @property
    def capacity(self):
        """The capacity (float, 0.0-1.0) of the active backing logical port."""
        return self._get_val_percent(_VNICD_DES_CAP_PCT)

    @property
    def auto_pri_failover(self):
        return self._get_val_bool(_VNICD_AUTO_FB)

    @auto_pri_failover.setter
    def auto_pri_failover(self, val):
        self.set_parm_value(_VNICD_AUTO_FB, u.sanitize_bool_for_api(val))


@ewrap.ElementWrapper.pvm_type(_VNICBD, has_metadata=True,
                               child_order=_VNICBD_EL_ORDER)
class VNICBackDev(ewrap.ElementWrapper):
    """SR-IOV backing device for a vNIC."""

    @classmethod
    def bld(cls, adapter, vios_uuid, sriov_adap_id, pport_id, capacity=None,
            failover_pri=None):
        """Create a new VNICBackDev, suitable for inclusion in a VNIC wrapper.

        :param adapter: pypowervm.adapter.Adapter for REST API communication.
        :param vios_uuid: String UUID of the Virtual I/O Server to host the
                          vNIC server for this backing device.
        :param sriov_adap_id: Integer SR-IOV Adapter ID of the SR-IOV adapter
                              owning the physical port on which the backing VF
                              is to be created: SRIOVAdapter.sriov_adap_id.
        :param pport_id: Integer physical port ID of SR-IOV physical port on
                         which the VF is to be created: SRIOVEthPPort.port_id
        :param capacity: Float value between 0.0 and 1.0 indicating the minimum
                         fraction of the physical port's bandwidth allocated to
                         traffic over this backing device.  Must be a multiple
                         of SRIOVEthPPort.min_granularity for the physical port
                         indicated by pport_id.  If not specified,
                         SRIOVEthPPort.min_granularity is used by the platform.
        :param failover_pri: Positive integer value representing the failover
                             priority of this backing device.
        :return: A new VNICBackDev, suitable for inclusion in a VNIC wrapper.
        """
        bdev = super(VNICBackDev, cls)._bld(adapter)
        # TODO(IBM): Verify that this can be ManagedSystem-less
        bdev._vios_href(adapter.build_href(_VIOS, vios_uuid, xag=[]))
        bdev._sriov_adap_id(sriov_adap_id)
        bdev._pport_id(pport_id)
        if capacity is not None:
            bdev._capacity(capacity)
        if failover_pri is not None:
            bdev.failover_pri = failover_pri
        return bdev

    @property
    def vios_href(self):
        return self.get_href(_VNICBD_VIOS, one_result=True)

    def _vios_href(self, href):
        self.set_href(_VNICBD_VIOS, href)

    @property
    def sriov_adap_id(self):
        return self._get_val_int(_VNICBD_SRIOV_ADP_ID)

    def _sriov_adap_id(self, val):
        self.set_parm_value(_VNICBD_SRIOV_ADP_ID, val)

    @property
    def pport_id(self):
        return self._get_val_int(_VNICBD_PPORT_ID)

    def _pport_id(self, val):
        self.set_parm_value(_VNICBD_PPORT_ID, val)

    @property
    def lport_href(self):
        return self.get_href(_VNICBD_LPORT, one_result=True)

    @property
    def capacity(self):
        """Gets the allocated capacity in a float-percentage format.

        :return: If the property is say "2.45%", a value of .0245 will be
                 returned.
        """
        return self._get_val_percent(_VNICBD_CUR_CAP_PCT)

    def _capacity(self, float_val):
        self.set_parm_value(_VNICBD_CUR_CAP_PCT,
                            u.sanitize_percent_for_api(float_val))

    @property
    def failover_pri(self):
        """The failover priority value for this backing device.

        :return: A value between 1 and 100, inclusive, with a lower number
                 indicating the higher priority (i.e. the backingdevice with
                 priority 1 will take precedence over that with priority 2).
        """
        return self._get_val_int(_VNICBD_FAILOVER_PRI)

    @failover_pri.setter
    def failover_pri(self, val):
        self.set_parm_value(_VNICBD_FAILOVER_PRI, val, attrib=pc.ATTR_KSV140)

    @property
    def is_active(self):
        return self._get_val_bool(_VNICBD_ACTIVE)

    @property
    def status(self):
        return self._get_val_str(_VNICBD_STATUS)


@ewrap.ElementWrapper.pvm_type(_IO_ADPT_CHOICE, has_metadata=False)
class LinkAggrIOAdapterChoice(ewrap.ElementWrapper):
    """A free I/O Adapter link aggregation choice.

    Flattens this two step hierarchy to pull the information needed directly
    from the IOAdapter element.
    """
    def __get_prop(self, func):
        """Thin wrapper to get the IOAdapter and get a property."""
        elem = self._find('IOAdapter')
        if elem is None:
            return None

        io_adpt = IOAdapter.wrap(elem)
        return getattr(io_adpt, func)

    @property
    def id(self):
        return self.__get_prop('id')

    @property
    def description(self):
        return self.__get_prop('description')

    @property
    def dev_name(self):
        return self.__get_prop('dev_name')

    @property
    def dev_type(self):
        return self.__get_prop('dev_type')

    @property
    def drc_name(self):
        return self.__get_prop('drc_name')

    @property
    def phys_loc_code(self):
        return self.__get_prop('phys_loc_code')

    @property
    def udid(self):
        return self.__get_prop('udid')



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tasks related to objects that will appear as hard disks to VMs."""

from pypowervm.tasks.hdisk import _fc
from pypowervm.tasks.hdisk import _iscsi

LUAType = _fc.LUAType
LUAStatus = _fc.LUAStatus
normalize_lun = _fc.normalize_lun
ITL = _fc.ITL
good_discovery = _fc.good_discovery
build_itls = _fc.build_itls
discover_hdisk = _fc.discover_hdisk
lua_recovery = _fc.lua_recovery
remove_hdisk = _fc.remove_hdisk
get_pg83_via_job = _fc.get_pg83_via_job
discover_iscsi = _iscsi.discover_iscsi
discover_iscsi_initiator = _iscsi.discover_iscsi_initiator
remove_iscsi = _iscsi.remove_iscsi



# Copyright 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import ast

from oslo_log import log as logging

import pypowervm.const as c
from pypowervm.i18n import _
from pypowervm.wrappers import job
from pypowervm.wrappers.virtual_io_server import VIOS

LOG = logging.getLogger(__name__)
_JOB_NAME = "ISCSIDiscovery"
_ISCSI_LOGOUT = "ISCSILogout"


class TransportType(object):
    """Valid values for iSCSI transport types."""
    ISCSI = 'iscsi'
    ISER = 'iser'


def discover_iscsi(adapter, host_ip, user, password, iqn, vios_uuid,
                   transport_type=None):
    """Runs iscsi discovery and login job

    :param adapter: pypowervm adapter
    :param host_ip: The ip address of the iscsi target.
    :param user: The username needed for authentication.
    :param password: The password needed for authentication.
    :param iqn: The IQN (iSCSI Qualified Name) of the created volume on the
                target. (e.g. iqn.2016-06.world.srv:target00)
    :param vios_uuid: The uuid of the VIOS (VIOS must be a Novalink VIOS type).
    :param transport_type: The type of the volume to be connected. Must be a
                           valid TransportType.
    :return: The device name of the created volume.
    :return: The UniqueDeviceId of the create volume.
    """

    resp = adapter.read(VIOS.schema_type, vios_uuid,
                        suffix_type=c.SUFFIX_TYPE_DO, suffix_parm=(_JOB_NAME))
    job_wrapper = job.Job.wrap(resp)

    # Create job parameters
    job_parms = [job_wrapper.create_job_parameter('hostIP', host_ip)]
    job_parms.append(job_wrapper.create_job_parameter('password', password))
    job_parms.append(job_wrapper.create_job_parameter('user', user))
    job_parms.append(job_wrapper.create_job_parameter('targetIQN', iqn))
    if transport_type is not None:
        job_parms.append(
            job_wrapper.create_job_parameter('transportType', transport_type))
    job_wrapper.run_job(vios_uuid, job_parms=job_parms, timeout=120)
    results = job_wrapper.get_job_results_as_dict()

    # DEV_OUTPUT: ["IQN1 dev1 udid", "IQN2 dev2 udid"]
    output = ast.literal_eval(results.get('DEV_OUTPUT'))
    # Find dev corresponding to given IQN
    for dev in output:
        if len(dev.split()) != 3:
            LOG.warning(_("Invalid device output: %(dev)s"), {'dev': dev})
            continue
        outiqn, outname, udid = dev.split()
        if outiqn == iqn:
            return outname, udid
    LOG.error(_("Expected IQN: %(IQN)s not found on iscsi target %(host_ip)s"),
              {'IQN': iqn, 'host_ip': host_ip})
    return None, None


def discover_iscsi_initiator(adapter, vios_uuid):
    """Discovers the initiator name.

    :param adapter: pypowervm adapter
    :param vios_uuid: The uuid of the VIOS (VIOS must be a Novalink VIOS type).
    :return: The iscsi initiator name.
    """
    resp = adapter.read(VIOS.schema_type, vios_uuid,
                        suffix_type=c.SUFFIX_TYPE_DO, suffix_parm=(_JOB_NAME))
    job_wrapper = job.Job.wrap(resp)

    job_wrapper.run_job(vios_uuid, timeout=120)
    results = job_wrapper.get_job_results_as_dict()

    # InitiatorName: iqn.2010-10.org.openstack:volume-4a75e9f7-dfa3
    return results.get('InitiatorName')


def remove_iscsi(adapter, targetIQN, vios_uuid):
    """Logout of an iSCSI session.

    The iSCSI volume with the given targetIQN must not have any mappings from
    the VIOS to a client when this is called.

    :param adapter: pypowervm adapter
    :param targetIQN: The IQN (iSCSI Qualified Name) of the created volume on
                      the target. (e.g. iqn.2016-06.world.srv:target00)
    :param vios_uuid: The uuid of the VIOS (VIOS must be a Novalink VIOS type).
    """
    resp = adapter.read(VIOS.schema_type, vios_uuid,
                        suffix_type=c.SUFFIX_TYPE_DO,
                        suffix_parm=(_ISCSI_LOGOUT))
    job_wrapper = job.Job.wrap(resp)

    job_parms = [job_wrapper.create_job_parameter('targetIQN', targetIQN)]
    job_wrapper.run_job(vios_uuid, job_parms=job_parms, timeout=120)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for the raw PHYP long term metrics."""

import testtools

from pypowervm.tests.test_utils import pvmhttp
from pypowervm.wrappers.pcm import phyp as pcm_phyp

PHYP_DATA = 'phyp_pcm_data.txt'


class TestPhypLTM(testtools.TestCase):

    def setUp(self):
        super(TestPhypLTM, self).setUp()

        self.raw_json = pvmhttp.PVMFile(PHYP_DATA).body

    def test_parse(self):
        info = pcm_phyp.PhypInfo(self.raw_json)
        self.assertIsNotNone(info)

        # Validate the info
        self.assertEqual('1.3.0', info.info.version)
        self.assertEqual('Raw', info.info.metric_type)
        self.assertEqual('LTM', info.info.monitoring_type)
        self.assertEqual('8247-22L*2125D4A', info.info.mtms)
        self.assertEqual('dev-4', info.info.name)

        # Validate some samples
        sample = info.sample
        self.assertEqual(806297258933150.0, sample.time_based_cycles)
        self.assertEqual(0, sample.status)
        self.assertEqual(u'2015-05-27T08:17:45+0000', sample.time_stamp)

        # Firmware
        self.assertEqual(58599310268,
                         sample.system_firmware.utilized_proc_cycles)
        self.assertEqual(4096, sample.system_firmware.assigned_mem)

        # Shared Proc Pool
        spp_list = sample.shared_proc_pools
        self.assertEqual(1, len(spp_list))
        self.assertEqual(0, spp_list[0].id)
        self.assertEqual('DefaultPool', spp_list[0].name)
        self.assertEqual(1.6125945162342e+16, spp_list[0].assigned_proc_cycles)
        self.assertEqual(683011326288, spp_list[0].utilized_pool_cycles)
        self.assertEqual(20, spp_list[0].max_proc_units)
        self.assertEqual(18, spp_list[0].borrowed_pool_proc_units)

        # Processor
        self.assertEqual(20, sample.processor.total_proc_units)
        self.assertEqual(20, sample.processor.configurable_proc_units)
        self.assertEqual(18.9, sample.processor.available_proc_units)
        self.assertEqual(512000000, sample.processor.proc_cycles_per_sec)

        # Memory
        self.assertEqual(65536, sample.memory.total_mem)
        self.assertEqual(32512, sample.memory.available_mem)
        self.assertEqual(65536, sample.memory.configurable_mem)

        # LPARs
        self.assertEqual(5, len(sample.lpars))

        # First LPAR shouldn't have network or storage (inactive)
        bad_lpar = sample.lpars[0]
        self.assertEqual(6, bad_lpar.id)
        self.assertEqual('2545BCC5-BAE8-4414-AD49-EAFC2DEE2546', bad_lpar.uuid)
        self.assertEqual('aixlinux', bad_lpar.type)
        self.assertEqual('fkh4-99b8fdca-kyleh', bad_lpar.name)
        self.assertEqual('Not Activated', bad_lpar.state)
        self.assertEqual(100, bad_lpar.affinity_score)
        self.assertIsNotNone(bad_lpar.memory)
        self.assertIsNotNone(bad_lpar.processor)
        self.assertEqual(None, bad_lpar.network)
        self.assertEqual(None, bad_lpar.storage)

        # Last LPAR should have network and storage
        good_lpar = sample.lpars[4]

        # VM Memory
        self.assertEqual(20480, good_lpar.memory.logical_mem)
        self.assertEqual(20480, good_lpar.memory.backed_physical_mem)

        # VM Processor
        self.assertEqual(0, good_lpar.processor.pool_id)
        self.assertEqual('uncap', good_lpar.processor.mode)
        self.assertEqual(4, good_lpar.processor.virt_procs)
        self.assertEqual(.4, good_lpar.processor.proc_units)
        self.assertEqual(128, good_lpar.processor.weight)
        self.assertEqual(1765629232513,
                         good_lpar.processor.entitled_proc_cycles)
        self.assertEqual(264619289721,
                         good_lpar.processor.util_cap_proc_cycles)
        self.assertEqual(641419282, good_lpar.processor.util_uncap_proc_cycles)
        self.assertEqual(0, good_lpar.processor.idle_proc_cycles)
        self.assertEqual(0, good_lpar.processor.donated_proc_cycles)
        self.assertEqual(0, good_lpar.processor.time_wait_dispatch)
        self.assertEqual(160866895489, good_lpar.processor.total_instructions)
        self.assertEqual(193139925064,
                         good_lpar.processor.total_inst_exec_time)

        # VM Vea
        vea = good_lpar.network.veas[0]
        self.assertEqual(2227, vea.vlan_id)
        self.assertEqual(0, vea.vswitch_id)
        self.assertEqual('U8247.22L.2125D4A-V2-C2', vea.physical_location)
        self.assertEqual(True, vea.is_pvid)
        self.assertEqual(10, vea.received_packets)
        self.assertEqual(100, vea.sent_packets)
        self.assertEqual(5, vea.dropped_packets)
        self.assertEqual(100, vea.sent_bytes)
        self.assertEqual(10000, vea.received_bytes)
        self.assertEqual(0, vea.received_physical_packets)
        self.assertEqual(0, vea.sent_physical_packets)
        self.assertEqual(0, vea.dropped_physical_packets)
        self.assertEqual(0, vea.sent_physical_bytes)
        self.assertEqual(0, vea.received_physical_bytes)

        # TODO(thorst) Test SR-IOV

        # VM storage
        stor = good_lpar.storage.v_stor_adpts[0]
        self.assertEqual('U8247.22L.2125D4A-V2-C3', stor.physical_location)
        self.assertEqual(1, stor.vios_id)
        self.assertEqual(1000, stor.vios_slot)

        # Test that VFC adapter has been parsed.
        self.assertIsNotNone(good_lpar.storage.v_fc_adpts)
        self.assertEqual(2, len(good_lpar.storage.v_fc_adpts))
        # Test 1st VFC adapter
        vfc_adpt = good_lpar.storage.v_fc_adpts[0]
        self.assertEqual('U8247.22L.2125D4A-V2-C2',
                         vfc_adpt.physical_location)
        self.assertEqual(2, vfc_adpt.vios_id)
        self.assertEqual(2, len(vfc_adpt.wwpn_pair))
        self.assertIn(13857705835384867080, vfc_adpt.wwpn_pair)
        self.assertIn(13857705835384867081, vfc_adpt.wwpn_pair)
        # Test 2nd VFC adapter
        vfc_adpt = good_lpar.storage.v_fc_adpts[1]
        self.assertEqual('U8247.22L.2125D4A-V2-C3',
                         vfc_adpt.physical_location)
        self.assertEqual(1, vfc_adpt.vios_id)
        self.assertEqual(2, len(vfc_adpt.wwpn_pair))
        self.assertIn(13857705835384867082, vfc_adpt.wwpn_pair)
        self.assertIn(13857705835384867083, vfc_adpt.wwpn_pair)

        # TODO(thorst) Test vfc



# Copyright 2015, 2016 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tasks specific to partitions (LPARs and VIOSes)."""

from oslo_log import log as logging
import time

import pypowervm.const as c
import pypowervm.exceptions as ex
from pypowervm.i18n import _
import pypowervm.util as u
import pypowervm.utils.transaction as tx
import pypowervm.wrappers.base_partition as bp
import pypowervm.wrappers.logical_partition as lpar
import pypowervm.wrappers.virtual_io_server as vios

LOG = logging.getLogger(__name__)

# RMC must be either active or busy.  Busy is allowed because that simply
# means that something is running against the VIOS at the moment...but
# it should recover shortly.
_VALID_RMC_STATES = (bp.RMCState.ACTIVE, bp.RMCState.BUSY)

# Only a running state is OK for now.
_VALID_VM_STATES = (bp.LPARState.RUNNING,)

# Not the opposite of the above
_DOWN_VM_STATES = (bp.LPARState.NOT_ACTIVATED, bp.LPARState.ERROR,
                   bp.LPARState.NOT_AVAILBLE, bp.LPARState.SHUTTING_DOWN,
                   bp.LPARState.SUSPENDED, bp.LPARState.SUSPENDING,
                   bp.LPARState.UNKNOWN)


def get_mgmt_partition(adapter):
    """Get the LPAR/VIOS wrapper representing the PowerVM management partition.

    :param adapter: The pypowervm.adapter.Adapter through which to query the
                    REST API.
    :return: pypowervm.wrappers.logical_partition.LPAR/virtual_io_server.VIOS
             wrapper representing the management partition.
    :raise ManagementPartitionNotFoundException: if we don't find exactly one
                                                 management partition.
    """

    # There will almost always be fewer VIOSes than LPARs.  Since we're
    # querying without xags, it should be very quick.
    vio_wraps = vios.VIOS.search(adapter, is_mgmt_partition=True)
    if len(vio_wraps) == 1:
        return vio_wraps[0]

    # We delay the query to the LPARs because there could be hundreds of them.
    # So we don't want to query it unless we need to.
    lpar_wraps = lpar.LPAR.search(adapter, is_mgmt_partition=True)
    if len(lpar_wraps) == 1:
        return lpar_wraps[0]

    # If we made it here, something is wrong.
    raise ex.ManagementPartitionNotFoundException(
        count=len(vio_wraps + lpar_wraps))


def get_this_partition(adapter):
    """Get the LPAR/VIOS wrapper of the node on which this method is running.

    :param adapter: The pypowervm.adapter.Adapter through which to query the
                    REST API.
    :return: pypowervm.wrappers.logical_partition.LPAR/virtual_io_server.VIOS
             wrapper representing the local partition.
    :raise LocalPartitionNotFoundException: if we don't find exactly one LPAR/
                                            VIOS with the local VM's short ID.
    """
    myid = u.my_partition_id()

    # There will almost always be fewer VIOSes than LPARs.  Since we're
    # querying without xags, it should be very quick.
    vio_wraps = vios.VIOS.search(adapter, id=myid)
    if len(vio_wraps) == 1:
        return vio_wraps[0]

    # We delay the query to the LPARs because there could be hundreds of them.
    # So we don't want to query it unless we need to.
    lpar_wraps = lpar.LPAR.search(adapter, id=myid)
    if len(lpar_wraps) == 1:
        return lpar_wraps[0]

    # If we made it here, something is wrong.
    raise ex.ThisPartitionNotFoundException(
        count=len(vio_wraps + lpar_wraps), lpar_id=myid)


def get_active_vioses(adapter, xag=(), vios_wraps=None, find_min=None):
    """Returns a list of active Virtual I/O Server Wrappers for a host.

    Active is defined by powered on and RMC state being 'active'.  The
    VIOSes will be sorted such that if the Mgmt partition is a VIOS, it is
    the first in the list.

    :param adapter: The pypowervm adapter for the query.
    :param xag: (Optional, Default: ()) Iterable of extended attributes to use.
    :param vios_wraps: (Optional, Default: None) A list of VIOS wrappers. If
                       specified, the method will check for active VIOSes
                       in this list instead of issuing a GET.
    :param find_min: (Optional, Default: None) If specified, the minimum
                     acceptable number of active VIOSes.  If fewer are found,
                     this method raises NotEnoughActiveVioses.
    :return: List of VIOS wrappers.
    :raise NotEnoughActiveVioses: If find_min is specified and the number of
                                  active VIOSes is less than the specified
                                  number.
    """
    if vios_wraps is None:
        vios_wraps = vios.VIOS.get(adapter, xag=xag)

    # A VIOS is 'active' if it is powered on and either RMC is active or it
    # is the mgmt partition.
    ret = [vio for vio in vios_wraps if vio.state in _VALID_VM_STATES and
           (vio.rmc_state in _VALID_RMC_STATES or vio.is_mgmt_partition)]
    ret = sorted(ret, key=lambda x: x.is_mgmt_partition, reverse=True)

    if find_min is not None and len(ret) < find_min:
        raise ex.NotEnoughActiveVioses(exp=find_min, act=len(ret))

    LOG.debug('Found active VIOS(es): %s', str([vio.name for vio in ret]))

    return ret


def get_partitions(adapter, lpars=True, vioses=True, mgmt=False):
    """Get a list of partitions.

    Can include LPARs, VIOSes, and the management partition.

    :param adapter: The pypowervm adapter.
    :param lpars: If True, the result will include all LPARs.
    :param vioses: If True, the result will include all VIOSes.
    :param mgmt: If True, the result is guaranteed to include the
                 management partition, even if it would not otherwise have
                 been included based on get_lpars/get_vioses.
    """
    rets = []
    if vioses:
        rets.extend(vios.VIOS.get(adapter))
    if lpars:
        rets.extend(lpar.LPAR.get(adapter))

    # If they need the mgmt lpar, get it.  But ONLY if we didn't get both
    # VIOSes and LPARs.  If we got both of those already, then we are
    # guaranteed to already have the mgmt lpar in there.
    if mgmt and not (lpars and vioses):
        mgmt_w = get_mgmt_partition(adapter)
        if mgmt_w.uuid not in [x.uuid for x in rets]:
            rets.append(get_mgmt_partition(adapter))

    return rets


def get_physical_wwpns(adapter):
    """Returns the active WWPNs of the FC ports across all VIOSes on system.

    :param adapter: pypowervm.adapter.Adapter for REST API communication.
    """
    vios_feed = vios.VIOS.get(adapter, xag=[c.XAG.VIO_STOR])
    wwpn_list = []
    for vwrap in vios_feed:
        wwpn_list.extend(vwrap.get_active_pfc_wwpns())
    return wwpn_list


def build_active_vio_feed_task(adapter, name='vio_feed_task', xag=(
        c.XAG.VIO_STOR, c.XAG.VIO_SMAP, c.XAG.VIO_FMAP)):
    """Builds a FeedTask for all active VIOSes.

    The transaction FeedTask enables users to collect a set of 'WrapperTasks'
    against a feed of entities (in this case a set of active VIOSes). The
    WrapperTask (within the FeedTask) handles lock and retry.

    This is useful to batch together a set of updates across a feed of elements
    (and multiple updates within a given wrapper).  This allows for significant
    performance improvements.

    :param adapter: The pypowervm adapter for the query.
    :param name: (Optional) The name of the feed manager.  Defaults to
                 vio_feed_task.
    :param xag: (Optional) Iterable of extended attributes to use.  If not
                specified, defaults to all mapping/storage options (as this is
                most common case for using a transaction manager).
    :raise NotEnoughActiveVioses: if there is not at least one active VIOS.
    """
    return tx.FeedTask(name, get_active_vioses(adapter, xag=xag, find_min=1))


def _rmc_down(vwrap):
    """Check if VIOS is in RMC Down state.

    :param vwrap: VIOS wrapper on which to check if RMC is down
    """
    if vwrap.is_mgmt_partition:
        return False
    if (vwrap.rmc_state not in _VALID_RMC_STATES and
            vwrap.state not in _DOWN_VM_STATES):
        return True
    return False


def _vios_waits_timed_out(no_rmc_vwraps, time_waited, max_wait_time=None):
    """Determine whether we've waited long enough for active VIOSes to get RMC.

    If max_wait_time is None, we will determine a suitable max_wait_time based
    on how long each VIOS has been booted.

    Then this method simply returns whether the time_waited exceeds the
    (specified or generated) max_wait_time for all VIOSes.

    :param no_rmc_vwraps: List of state-up/RMC-down VIOS wrappers.
    :param time_waited: The number of seconds the caller has waited thus far.
    :param max_wait_time: The maximum total number of seconds we should wait
                          before declaring we've waited long enough.  If None,
                          a suitable value will be determined based on the
                          VIOS's uptime.
    :return: True if we've waited long enough for this VIOS to come up.  False
             if we should wait some more.
    """
    for vwrap in no_rmc_vwraps:
        if max_wait_time is None:
            wait_time = 120 if vwrap.uptime > 3600 else 600
        else:
            wait_time = max_wait_time
        if time_waited < wait_time:
            # If we haven't waited long enough for *any* VIOS, keep waiting.
            return False
    return True


def _wait_for_vioses(adapter, max_wait_time=None):
    """Wait for VIOSes to stabilize, and report on their states.

    :param adapter: The pypowervm adapter for the query.
    :param max_wait_time: Maximum number of seconds to wait for running VIOSes
                          to get an active RMC connection.  If None, we will
                          wait longer if the VIOS booted recently, shorter if
                          it has been up for a while.
    :return: List of all VIOSes returned by the REST API.
    :return: List of all VIOSes which are powered on, but with RMC inactive.
    """
    vios_wraps = []
    rmc_down_vioses = []
    sleep_step = 5
    time_waited = 0
    while True:
        try:
            vios_wraps = vios.VIOS.get(adapter)
            rmc_down_vioses = [
                vwrap for vwrap in vios_wraps if _rmc_down(vwrap)]
            if not vios_wraps or (not rmc_down_vioses and get_active_vioses(
                    adapter, vios_wraps=vios_wraps)):
                # If there are truly no VIOSes (which should generally be
                # impossible if this code is running), we'll fail.
                # If at least one VIOS is up, and all active VIOSes have RMC,
                # we'll succeed.
                break
        except Exception as e:
            # Things like "Service Unavailable"
            LOG.warning(e)
        # If we get here, we're only waiting for VIOSes that are
        # state-active/RMC-down.  On each iteration, if a new VIOS comes up, it
        # will be considered here until its RMC comes up.
        if rmc_down_vioses and _vios_waits_timed_out(rmc_down_vioses,
                                                     time_waited,
                                                     max_wait_time):
            break
        time.sleep(sleep_step)
        time_waited += sleep_step
    return vios_wraps, rmc_down_vioses, time_waited


def validate_vios_ready(adapter, max_wait_time=None):
    """Check whether VIOS rmc is up and running on this host.

    Will query the VIOSes for a period of time attempting to ensure all
    running VIOSes get an active RMC.  If no VIOSes are ready by the timeout,
    ViosNotAvailable is raised.  If only some of the VIOSes had RMC go active
    by the end of the wait period, the method will complete.

    :param adapter: The pypowervm adapter for the query.
    :param max_wait_time: Integer maximum number of seconds to wait for running
                          VIOSes to get an active RMC connection.  Defaults to
                          None, in which case the system will determine an
                          appropriate amount of time to wait.  This can be
                          influenced by whether or not the VIOS just booted.
    :raises: A ViosNotAvailable exception if a VIOS is not available by a
             given timeout.
    """
    # Used to keep track of VIOSes and reduce queries to API
    vwraps, rmc_down_vioses, waited = _wait_for_vioses(adapter, max_wait_time)

    if rmc_down_vioses:
        LOG.warning(
            _('Timed out waiting for the RMC state of all the powered on '
              'Virtual I/O Servers to be active. Wait time was: %(time)d '
              'seconds. VIOSes that did not go active were: %(vioses)s.'),
            {'time': waited,
             'vioses': ', '.join([vio.name for vio in rmc_down_vioses])})

    # If we didn't get a single active VIOS then raise an exception
    if not get_active_vioses(adapter, vios_wraps=vwraps):
        raise ex.ViosNotAvailable(wait_time=waited)


def has_physical_io(part_w):
    """Determine whether a partition has any physical I/O adapters attached.

    This method looks over all of the slots on the partition and returns True
    if any of the slots has something in it, except for the following:
        - USB Device
        - Graphics
        - Empty Slot
    All other devices are considered I/O.

    :param part_w: Wrapper (LPAR or VIOS) of the partition to check.
    :return: True if any physical I/O adapter was found attached to the
             partition; False otherwise.
    """
    try:
        io_slots = part_w.io_config.io_slots
    except AttributeError:
        # If the wrapper has no io_slots, it has no physical I/O
        return False

    # Doesn't count as physical I/O if description contains any of these
    non_ios = ("USB", "Universal Serial Bus", "Graphics", "Empty slot")
    for io_slot in io_slots:
        try:
            # If the description *isn't* one of the non-I/O ones, it's a hit.
            if not any([non_io in io_slot.description for non_io in non_ios]):
                return True
        except AttributeError:
            # The slot didn't have a description.  That shouldn't happen.  But
            # we have to assume it's physical I/O in that case.
            LOG.warning(
                _("Assuming description-less slot is physical I/O: %s"),
                io_slot.toxmlstring())
            return True
    # We got through all the I/O slots without finding a physical I/O adapter
    return False



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities around Universally-Unique Identifiers (UUIDs)."""

import re

from pypowervm import const


def convert_uuid_to_pvm(uuid):
    """Converts a standard UUID to PowerVM format

    PowerVM uuids always set the byte 0, bit 0 to 0.

    :param uuid: A standard format uuid string
    :returns: A PowerVM compliant uuid
    """
    return "%x%s" % (int(uuid[0], 16) & 7, uuid[1:])


def id_or_uuid(an_id):
    """Sanitizes a short ID or string UUID, and indicates which was used.

    Use as:
        is_uuid, lpar_id = id_or_uuid(lpar_id)
        if is_uuid:  # lpar_id is a string UUID
        else:  # lpar_id is LPAR short ID of type int

    :param an_id: Short ID (may be string or int) or string UUID of, e.g., an
                  LPAR.
    :return: Boolean.  If True, the other return is a UUID string.  If False,
                       it is an integer.
    :return: The input ID, either converted to int, or in its original string
             form if a UUID.
    """
    if isinstance(an_id, str) and re.match(const.UUID_REGEX_WORD, an_id):
        is_uuid = True
        ret_id = an_id
    else:
        is_uuid = False
        ret_id = int(an_id)
    return is_uuid, ret_id



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tasks around ClientNetworkAdapter."""
from oslo_concurrency import lockutils

from pypowervm import exceptions as exc
from pypowervm.i18n import _
from pypowervm.tasks import partition
from pypowervm.wrappers import logical_partition as lpar
from pypowervm.wrappers import managed_system as pvm_ms
from pypowervm.wrappers import network as pvm_net

VLAN_LOCK = "reserve_vlan"


def crt_cna(adapter, host_uuid, lpar_uuid, pvid,
            vswitch=pvm_net.VSW_DEFAULT_VSWITCH, crt_vswitch=False,
            slot_num=None, mac_addr=None, addl_vlans=None):
    """Puts a new ClientNetworkAdapter on a given LPAR.

    This will update the LPAR and put a new CNA on it.  If the LPAR is active
    can only perform if there is an active RMC connection.  If the LPAR is
    powered off, then it will update it offline.

    :param adapter: The pypowervm adapter to perform the update through.
    :param host_uuid: The host UUID that the CNA will be put on.
    :param lpar_uuid: The lpar UUID to update.
    :param pvid: The primary VLAN ID.
    :param vswitch: The name of the virtual switch that this CNA will be
                    attached to.
    :param crt_vswitch: A boolean to indicate that if the vSwitch can not be
                        found, the system should attempt to create one (with
                        the default parameters - ex: Veb mode).
    :param slot_num: Optional slot number to use for the CNA.  If not
                     specified, will utilize the next available slot on the
                     LPAR.
    :param mac_addr: The optional mac address.  If not specified, one will be
                     auto generated.
    :param addl_vlans: Optional list of (up to 18) additional VLANs.  Can be
                       a list of Ints or Strings (that parse down to ints).
    :return: The CNA Wrapper that was created.
    """
    # Join the additional VLANs
    addl_tagged_vlans = None
    if addl_vlans is not None:
        addl_tagged_vlans = " ".join(addl_vlans)

    # Sanitize the pvid
    pvid = str(pvid)

    # Find the appropriate virtual switch.
    vswitch_w = _find_or_create_vswitch(adapter, host_uuid, vswitch,
                                        crt_vswitch)

    # Find the virtual network.  Ensures that the system is ready for this.
    if adapter.traits.vnet_aware:
        _find_or_create_vnet(adapter, host_uuid, pvid, vswitch_w)

    # Build and create the CNA
    net_adpt = pvm_net.CNA.bld(
        adapter, pvid, vswitch_w.related_href, slot_num=slot_num,
        mac_addr=mac_addr, addl_tagged_vlans=addl_tagged_vlans)
    return net_adpt.create(parent_type=lpar.LPAR, parent_uuid=lpar_uuid)


def _find_or_create_vnet(adapter, host_uuid, vlan, vswitch):
    # Read the existing virtual networks.  Try to locate...
    vnet_feed_resp = adapter.read(pvm_ms.System.schema_type, host_uuid,
                                  pvm_net.VNet.schema_type)
    vnets = pvm_net.VNet.wrap(vnet_feed_resp)
    for vnet in vnets:
        if vlan == str(vnet.vlan) and vnet.vswitch_id == vswitch.switch_id:
            return vnet

    # Must not have found it.  Lets try to create it.
    name = '%(vswitch)s-%(vlan)s' % {'vswitch': vswitch.name,
                                     'vlan': str(vlan)}
    # VLAN 1 is not allowed to be tagged.  All others are.  VLAN 1 would be
    # used for 'Flat' networks most likely.
    tagged = (vlan != '1')
    vnet = pvm_net.VNet.bld(adapter, name, vlan, vswitch.related_href, tagged)
    return vnet.create(parent_type=pvm_ms.System, parent_uuid=host_uuid)


def _find_or_create_vswitch(adapter, host_uuid, vs_name, crt_vswitch):
    """Finds (or creates) the appropriate virtual switch.

    :param adapter: The pypowervm adapter to perform the update through.
    :param host_uuid: The host UUID that the CNA will be put on.
    :param vs_name: The name of the virtual switch that this CNA will be
                    attached to.
    :param crt_vswitch: A boolean to indicate that if the vSwitch can not be
                        found, the system should attempt to create one (with
                        the default parameters - ex: Veb mode).
    """
    vswitch_w = pvm_net.VSwitch.search(adapter, parent_type=pvm_ms.System,
                                       parent_uuid=host_uuid, one_result=True,
                                       name=vs_name)

    if vswitch_w is None:
        if crt_vswitch:
            vswitch_w = pvm_net.VSwitch.bld(adapter, vs_name)
            vswitch_w = vswitch_w.create(parent_type=pvm_ms.System,
                                         parent_uuid=host_uuid)
        else:
            raise exc.Error(_('Unable to find the Virtual Switch %s on the '
                              'system.') % vs_name)
    return vswitch_w


def _find_free_vlan(adapter, host_uuid, vswitch_w):
    """Finds a free VLAN on the vswitch specified."""

    # A Virtual Network (VNet) will exist for every PowerVM vSwitch / VLAN
    # combination in the system.  Getting the feed is a quick way to determine
    # which VLANs are in use.
    vnet_resp_feed = adapter.read(pvm_ms.System.schema_type,
                                  root_id=host_uuid,
                                  child_type=pvm_net.VNet.schema_type)
    vnets = pvm_net.VNet.wrap(vnet_resp_feed)

    # Use that feed to get the VLANs in use, but only get the ones in use for
    # the vSwitch passed in.
    used_vids = [x.vlan for x in vnets
                 if x.associated_switch_uri == vswitch_w.related_href]

    # Walk through the VLAN range, and as soon as one is found that is not in
    # use, return it to the user.
    for x in range(1, 4094):
        if x not in used_vids:
            return x

    raise exc.Error(_('Unable to find a valid VLAN for Virtual Switch %s.') %
                    vswitch_w.name)


@lockutils.synchronized(VLAN_LOCK)
def assign_free_vlan(adapter, host_uuid, vswitch_w, cna, ensure_enabled=False):
    """Assigns a free vlan to a given cna. Also ensure the CNA is enabled.

    :param adapter: The adapter to read the vnet information from
    :param host_uuid: The host UUID that the CNA is on
    :param vswitch_w: The vswitch wrapper to find the free vlan on.
    :param cna: The CNA wrapper to be updated with a new vlan.
    :param ensure_enabled: (Optional, Default: False) If true, enable the CNA
                           before updating.
    :return: The updated CNA.
    """

    vlan = _find_free_vlan(adapter, host_uuid, vswitch_w)
    cna.pvid = vlan
    if ensure_enabled:
        cna.enabled = True
    cna = cna.update()
    return cna


@lockutils.synchronized(VLAN_LOCK)
def crt_trunk_with_free_vlan(
        adapter, host_uuid, src_io_host_uuids, vs_name,
        crt_vswitch=True, dev_name=None):
    """Creates a trunk adapter(s) with a free VLAN on the system.

    :param adapter: The pypowervm adapter to perform the update through.
    :param host_uuid: The host UUID that the CNA will be put on.
    :param src_io_host_uuids: The list of UUIDs of the LPARs that will host the
                              Trunk Adapters.  At least one UUID is required.
                              Multiple will be supported, and the Trunk
                              Priority will increment per adapter (in the order
                              that the I/O hosts are specified).
    :param pvid: The port VLAN ID.
    :param vs_name: The name of the PowerVM Hypervisor Virtual Switch to create
                    the p2p connection on.  This is required because it is not
                    recommended to create it on the default (ETHERNET0) virtual
                    switch.
    :param crt_vswitch: (Optional, Default: True) A boolean to indicate that
                        if the vSwitch can not be found, the system should
                        attempt to create one (with the default parameters -
                        ex: Veb mode).
    :param dev_name: (Optional, Default: None) The device name.  Only valid
                     if the src_io_host_uuids is a single entity and the
                     uuid matches the mgmt lpar UUID.  Otherwise leave as
                     None.  If set, the name of the trunk adapter created on
                     the mgmt lpar will be set to this value.
    :return: The CNA Wrapper that was created.
    :return: The TrunkAdapters that were created.  Match the order that the
             src_io_host_uuids were passed in.
    """
    # Make sure we have the appropriate vSwitch
    vswitch_w = _find_or_create_vswitch(adapter, host_uuid, vs_name,
                                        crt_vswitch)

    # Find the free VLAN
    vlan = _find_free_vlan(adapter, host_uuid, vswitch_w)

    # Need to get the VIOS uuids to determine if the src_io_host_uuid is a VIOS
    iohost_wraps = partition.get_partitions(
        adapter, lpars=False, vioses=True, mgmt=True)
    io_uuid_to_wrap = {w.uuid: w for w in iohost_wraps
                       if w.uuid in src_io_host_uuids}

    # Now create the corresponding Trunk
    trunk_adpts = []
    trunk_pri = 1
    for io_uuid in src_io_host_uuids:
        trunk_adpt = pvm_net.CNA.bld(
            adapter, vlan, vswitch_w.related_href, trunk_pri=trunk_pri,
            dev_name=dev_name)
        trunk_adpts.append(trunk_adpt.create(parent=io_uuid_to_wrap[io_uuid]))
        trunk_pri += 1
    return trunk_adpts


def crt_p2p_cna(adapter, host_uuid, lpar_uuid, src_io_host_uuids, vs_name,
                crt_vswitch=True, mac_addr=None, slot_num=None, dev_name=None):
    """Creates a 'point-to-point' Client Network Adapter.

    A point to point connection is one that has a VLAN that is shared only
    between the lpar and the appropriate trunk adapter(s).  The system will
    determine what a free VLAN is on the virtual switch and use that for the
    point to point connection.

    The method will return the Client Network Adapter and the corresponding
    Trunk Adapter that it has created.  There may be multiple Trunk Adapters
    created if multiple src_io_host_uuids are passed in.  The Trunk Adapters
    can be created against the Virtual I/O Servers or the NovaLink partition.

    Nothing prevents the system from allowing another Client Network Adapter
    from being created and attaching to the connection.  The point-to-point
    connection is only guaranteed at the point in time at which it was created.

    NOTE: See the note in src_io_host_uuids.  Currently this API will only
    support the NovaLink partition.  Others will be added.  This parameter is
    there for future facing compatibility.

    :param adapter: The pypowervm adapter to perform the update through.
    :param host_uuid: The host UUID that the CNA will be put on.
    :param lpar_uuid: The lpar UUID to update.
    :param src_io_host_uuids: The list of UUIDs of the LPARs that will host the
                              Trunk Adapters.  At least one UUID is required.
                              Multiple will be supported, and the Trunk
                              Priority will increment per adapter (in the order
                              that the I/O hosts are specified).
    :param pvid: The primary VLAN ID.
    :param vs_name: The name of the PowerVM Hypervisor Virtual Switch to create
                    the p2p connection on.  This is required because it is not
                    recommended to create it on the default (ETHERNET0) virtual
                    switch.
    :param crt_vswitch: (Optional, Default: True) A boolean to indicate that
                        if the vSwitch can not be found, the system should
                        attempt to create one (with the default parameters -
                        ex: Veb mode).
    :param mac_addr: (Optional, Default: None) The mac address.  If not
                     specified, one will be auto generated.
    :param slot_num: (Optional, Default: None) The slot number to use for the
                     CNA. If not specified, will utilize the next available
                     slot on the LPAR.
    :param dev_name: (Optional, Default: None) The device name.  Only valid
                     if the src_io_host_uuids is a single entity and the
                     uuid matches the mgmt lpar UUID.  Otherwise leave as
                     None.  If set, the trunk adapter created on the mgmt lpar
                     will be set to this value.
    :return: The CNA Wrapper that was created.
    :return: The TrunkAdapters that were created.  Match the order that the
             src_io_host_uuids were passed in.
    """

    trunk_adpts = crt_trunk_with_free_vlan(
        adapter, host_uuid, src_io_host_uuids, vs_name,
        crt_vswitch=crt_vswitch, dev_name=dev_name)

    # Darn lack of re-entrant locks
    with lockutils.lock(VLAN_LOCK):
        vswitch_w = _find_or_create_vswitch(adapter, host_uuid, vs_name,
                                            crt_vswitch)
        client_adpt = pvm_net.CNA.bld(
            adapter, trunk_adpts[0].pvid, vswitch_w.related_href,
            slot_num=slot_num, mac_addr=mac_addr)
        client_adpt = client_adpt.create(parent_type=lpar.LPAR,
                                         parent_uuid=lpar_uuid)

    return client_adpt, trunk_adpts


def find_trunks(adapter, cna_w):
    """Returns the Trunk Adapters associated with the CNA.

    :param adapter: The pypowervm adapter to perform the search with.
    :param cna_w: The Client Network Adapter to find the Trunk Adapters for.
    :return: A list of Trunk Adapters (sorted by Trunk Priority) that host
             the Client Network Adapter.
    """
    # VIOS and Management Partitions can host Trunk Adapters.
    host_wraps = partition.get_partitions(
        adapter, lpars=False, vioses=True, mgmt=True)

    # Find the corresponding trunk adapters.
    trunk_list = []
    for host_wrap in host_wraps:
        trunk = _find_trunk_on_lpar(adapter, host_wrap, cna_w)
        if trunk:
            trunk_list.append(trunk)

    # Sort by the trunk priority
    trunk_list.sort(key=lambda x: x.trunk_pri)
    return trunk_list


def _find_trunk_on_lpar(adapter, parent_wrap, client_vea):

    cna_wraps = pvm_net.CNA.get(adapter, parent=parent_wrap)
    for cna in cna_wraps:
        if (cna.is_trunk and cna.pvid == client_vea.pvid and
                cna.vswitch_id == client_vea.vswitch_id):
            return cna
    return None


def _find_all_trunks_on_lpar(adapter, parent_wrap, vswitch_id=None):
    """Returns all trunk adapters on a given vswitch.

    :param adapter: The pypowervm adapter to perform the search with.
    :param vswitch_id: The id of the vswitch to search for orphaned trunks
                       on.
    :return: A list of trunk adapters that are associated with the given
             vswitch_id.
    """
    cna_wraps = pvm_net.CNA.get(adapter, parent=parent_wrap)
    trunk_list = []
    for cna in cna_wraps:
        if (cna.is_trunk and (vswitch_id is None
                              or cna.vswitch_id == vswitch_id)):
            trunk_list.append(cna)
    return trunk_list


def _find_cna_wraps(adapter, vswitch_id=None):
    """Returns all CNAs.

    :param adapter: The pypowervm adapter to perform the search with.
    :param vswitch_id: This param is optional. If specified, the method will
                       only return CNAs associated with the given vswitch.
    :return: A list of CNAs that are optionally associated with the given
             vswitch_id.
    """
    # All lpars should be searched, including VIOSes
    lpar_wraps = partition.get_partitions(adapter)

    cna_wraps = []
    filtered_cna_wraps = []
    for lpar_wrap in lpar_wraps:
        cna_wraps.extend(pvm_net.CNA.get(adapter, parent=lpar_wrap))

    # If a vswitch_id is passed in then filter to only cnas on that vswitch
    if (vswitch_id):
        for cna in cna_wraps:
            if(cna.vswitch_id == vswitch_id):
                filtered_cna_wraps.append(cna)
        cna_wraps = filtered_cna_wraps
    return cna_wraps


def find_cnas_on_trunk(trunk_w, cna_wraps=None):
    """Returns the CNAs associated with the Trunk Adapter.

    :param trunk_w: The Trunk Adapter to find the Client Network Adapters for.
    :param cna_wraps: Optional param for passing in the list of CNA wraps
                      to search.  If the list is none, queries will be done
                      to build the list.
    :return: A list of Client Network Adapters that are hosted by the Trunk
             Adapter.
    """
    adapter = trunk_w.adapter

    # Find all the CNAs on the system
    if cna_wraps is None:
        cna_wraps = _find_cna_wraps(adapter)

    # Search the CNA wraps for matching CNAs
    cna_list = []
    for cna in cna_wraps:
        if ((not cna.uuid == trunk_w.uuid) and cna.pvid == trunk_w.pvid and
                cna.vswitch_id == trunk_w.vswitch_id):
            cna_list.append(cna)

    return cna_list


def find_orphaned_trunks(adapter, vswitch_name):
    """Returns all orphaned trunk adapters on a given vswitch.

    An orphaned trunk is a trunk adapter that does not have any associated
    CNAs.

    :param adapter: The pypowervm adapter to perform the search with.
    :param vswitch_name: The name of the vswitch to search for orphaned trunks
                         on.
    :return: A list of trunk adapters that do not have any associated CNAs
    """
    vswitch = pvm_net.VSwitch.search(
        adapter, parent_type=pvm_ms.System, one_result=True,
        name=vswitch_name)

    # May occur if the system does not host the vswitch passed in.
    if vswitch is None:
        return []
    vswitch_id = vswitch.switch_id

    # VIOS and Management Partitions can host Trunk Adapters.
    host_wraps = partition.get_partitions(
        adapter, lpars=False, vioses=True, mgmt=True)

    # Get all the CNA wraps on the vswitch
    cna_wraps = _find_cna_wraps(adapter, vswitch_id=vswitch_id)

    # Find all trunk adapters on the vswitch.
    trunk_list = []
    for host_wrap in host_wraps:
        trunks = _find_all_trunks_on_lpar(adapter, parent_wrap=host_wrap,
                                          vswitch_id=vswitch_id)
        trunk_list.extend(trunks)

    # Check if the trunk adapters are orphans
    orphaned_trunk_list = []
    for trunk in trunk_list:
        if not find_cnas_on_trunk(trunk, cna_wraps=cna_wraps):
            orphaned_trunk_list.append(trunk)

    return orphaned_trunk_list



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import pypowervm.const as pc
import pypowervm.tests.test_utils.test_wrapper_abc as twrap
import pypowervm.wrappers.vios_file as vf


class TestVIOSFile(twrap.TestWrapper):

    file = 'file_feed.txt'
    wrapper_class_to_test = vf.File

    def test_wrapper_class(self):
        self.assertEqual(vf.File.schema_type, 'File')
        self.assertEqual(vf.File.schema_ns, pc.WEB_NS)
        self.assertTrue(vf.File.has_metadata)
        self.assertEqual(vf.File.default_attrib, pc.DEFAULT_SCHEMA_ATTR)

    def test_file(self):
        self.assertTrue(len(self.entries) > 0)

        vio_file = self.entries[0]
        self.assertEqual(vio_file.schema_type, 'File')
        self.assertEqual('boot_9699a0f5', vio_file.file_name)
        self.assertEqual('1421736166276', vio_file.date_modified)
        self.assertEqual('application/octet-stream',
                         vio_file.internet_media_type)
        self.assertEqual('5cd8e4b0-083e-4c71-bcff-2432807cfdcc',
                         vio_file.file_uuid)
        self.assertEqual(25165824, vio_file.expected_file_size)
        self.assertEqual(25165824, vio_file.current_file_size)
        self.assertEqual(vf.FileType.DISK_IMAGE, vio_file.enum_type)
        self.assertEqual('14B854F7-42CE-4FF0-BD57-1D117054E701',
                         vio_file.vios_uuid)
        self.assertEqual('0300f8d6de00004b000000014a54555cd9.28',
                         vio_file.tdev_udid)



# Copyright 2015 IBM Corp.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tasks around VIOS-backed 'physical' fibre channel disks."""

import itertools

from lxml import etree
from oslo_log import log as logging

from pypowervm import const as c
import pypowervm.entities as ent
import pypowervm.exceptions as pexc
from pypowervm.i18n import _
import pypowervm.tasks.storage as tsk_stg
import pypowervm.utils.transaction as tx
from pypowervm.wrappers import job as pvm_job
from pypowervm.wrappers import virtual_io_server as pvm_vios

LOG = logging.getLogger(__name__)

_LUA_CMD_VERSION = '3'
_LUA_VERSION = '2.0'
_LUA_RECOVERY = 'LUARecovery'
_RM_HDISK = 'RemoveDevice'

_MGT_CONSOLE = 'ManagementConsole'


class LUAType(object):
    """LUA Vendors."""
    IBM = "IBM"
    EMC = "EMC"
    NETAPP = "NETAPP"
    HDS = "HDS"
    HP = "HP"
    OTHER = "OTHER"


class LUAStatus(object):
    """LUA Recovery status codes."""
    DEVICE_IN_USE = '1'
    ITL_NOT_RELIABLE = '2'
    DEVICE_AVAILABLE = '3'
    STORAGE_NOT_INTEREST = '4'
    LUA_NOT_INTEREST = '5'
    INCORRECT_ITL = '6'
    FOUND_DEVICE_UNKNOWN_UDID = '7'
    FOUND_ITL_ERR = '8'


def normalize_lun(scsi_id):
    """Normalize the lun id to Big Endian

    :param scsi_id: Volume lun id
    :return: Converted LUN id in Big Endian as per the RFC 4455
    """
    # PowerVM keeps LUN identifiers in hex format.
    lun = '%x' % int(scsi_id)
    # For drivers which support complex LUA lun-id exceeding more than 2
    # bytes in such cases we need to append 8 zeros else 12 zeros to
    # pass 8 byte lun-id
    if len(lun) == 8:
        lun += "00000000"
    else:
        lun += "000000000000"

    return lun


class ITL(object):
    """The Nexus ITL.

    See SCSI ITL.  This is the grouping of the SCSI initiator, target and
    LUN.
    """

    def __init__(self, initiator, target, lun):
        """Create the ITL.

        :param initiator: The initiator WWPN.
        :param target: The target WWPN.
        :param lun: The LUN identifier.  Ex. 2 (an int).  The identifier will
                    be formatted from a generic integer LUN ID to match
                    PowerVM's LUN Identifier format.
        """
        self.initiator = initiator.lower().replace(':', '')
        self.target = target.lower().replace(':', '')
        self.lun = normalize_lun(lun)

    def __eq__(self, other):
        if other is None or not isinstance(other, ITL):
            return False

        return (self.initiator == other.initiator and
                self.target == other.target and
                self.lun == other.lun)

    def __hash__(self):
        return hash(self.initiator) ^ hash(self.target) ^ hash(self.lun)

    def __ne__(self, other):
        return not self.__eq__(other)


def good_discovery(status, device_name):
    """Checks the hdisk discovery results for a good discovery.

    Acceptable LUA discovery statuses are :-
    DEVICE_AVAILABLE: hdisk discovered on all the ITL paths and available.
    DEVICE_IN_USE: hdisk discovered on all the ITL paths and is in-use by
    the server.
    FOUND_ITL_ERR: hdisk is discovered on some of the ITL paths and available.
    This can happen if there are multiple ITL nexus paths are passed, and
    hdisk is discovered on few of the paths only. This can happen if multiple
    target wwpns and vios wwpns exists and only few are connected. If hdisk
    can be discovered on ANY of the paths its considered for good discovery.
   """
    return device_name is not None and status in [
        LUAStatus.DEVICE_AVAILABLE, LUAStatus.DEVICE_IN_USE,
        LUAStatus.FOUND_ITL_ERR]


def build_itls(i_wwpns, t_wwpns, lun):
    """This method builds the list of ITLs for all of the permutations.

    An ITL is specific to an initiator, target, and LUN.  However, with multi
    pathing, there are several scenarios where a given LUN will have many ITLs
    because of multiple initiators or targets.

    The initiators should be tied to a given Virtual I/O Server (or perhaps
    specific WWPNs within a VIOS).

    :param i_wwpns: List or set of initiator WWPNs.
    :param t_wwpns: List or set of target WWPNs.
    :param lun: The LUN identifier.  Ex. 2 (an int).  The identifier will be
                formatted from a generic integer LUN ID to match PowerVM's
                LUN Identifier format.
    :return: List of all the ITL permutations.
    """
    return [ITL(i, t, lun) for i, t in itertools.product(i_wwpns, t_wwpns)]


def discover_hdisk(adapter, vios_uuid, itls, vendor=LUAType.OTHER):
    """Attempt to discover a hard disk attached to a Virtual I/O Server.

    See lua_recovery.  This method attempts that call and analyzes the
    results.  On certain failure conditions (see below), this method will find
    stale LPARs, scrub storage artifacts associated with them, and then retry
    lua_recovery.  The retry is only attempted once; that result is returned
    regardless.

    The main objective of this method is to resolve errors resulting from
    incomplete cleanup of previous LPARs.  The stale LPAR's storage mappings
    can cause hdisk discovery to fail because it thinks the hdisk is already in
    use.

    Retry conditions: The scrub-and-retry will be triggered if:
    o dev_name is None; or
    o status is anything other than DEVICE_AVAILABLE or FOUND_ITL_ERR.  (The
      latter is acceptable because it means we discovered some, but not all, of
      the ITLs.  This is okay as long as dev_name is set.)

    :param adapter: The pypowervm adapter.
    :param vios_uuid: The Virtual I/O Server UUID.
    :param itls: A list of ITL objects.
    :param vendor: The vendor for the LUN.  See the LUAType.* constants.
    :return status: The status code from the discover process.
                    See LUAStatus.* constants.
    :return dev_name: The name of the discovered hdisk.
    :return udid: The UDID of the device.
    """
    # First attempt
    status, devname, udid = lua_recovery(adapter, vios_uuid, itls,
                                         vendor=vendor)
    # Do we need to scrub and retry?
    if not good_discovery(status, devname):
        vwrap = pvm_vios.VIOS.get(adapter, uuid=vios_uuid,
                                  xag=(c.XAG.VIO_SMAP, c.XAG.VIO_FMAP))

        scrub_ids = tsk_stg.find_stale_lpars(vwrap)
        if scrub_ids:
            # Detailed warning message by _log_lua_status
            LOG.warning(_("hdisk discovery failed; will scrub stale storage "
                          "for LPAR IDs %s and retry."), scrub_ids)
            # Scrub from just the VIOS in question.
            scrub_task = tx.FeedTask('scrub_vios_%s' % vios_uuid, [vwrap])
            tsk_stg.add_lpar_storage_scrub_tasks(scrub_ids, scrub_task)
            scrub_task.execute()
            status, devname, udid = lua_recovery(adapter, vios_uuid, itls,
                                                 vendor=vendor)
    return status, devname, udid


def lua_recovery(adapter, vios_uuid, itls, vendor=LUAType.OTHER):
    """Logical Unit Address Recovery - discovery of a FC-attached hdisk.

    When a new disk is created externally (say on a block device), the Virtual
    I/O Server may or may not discover it immediately.  This method forces a
    discovery on a given Virtual I/O Server.

    :param adapter: The pypowervm adapter.
    :param vios_uuid: The Virtual I/O Server UUID.
    :param itls: A list of ITL objects.
    :param vendor: The vendor for the LUN.  See the LUAType.* constants.
    :return status: The status code from the discover process.
                    See LUAStatus.* constants.
    :return dev_name: The name of the discovered hdisk.
    :return udid: The UDID of the device.
    """
    # Reduce the ITLs to ensure no duplicates
    itls = set(itls)

    # Build the LUA recovery XML
    lua_xml = _lua_recovery_xml(itls, adapter, vendor=vendor)

    # Build up the job & invoke
    resp = adapter.read(
        pvm_vios.VIOS.schema_type, root_id=vios_uuid,
        suffix_type=c.SUFFIX_TYPE_DO, suffix_parm=_LUA_RECOVERY)
    job_wrapper = pvm_job.Job.wrap(resp)
    job_parms = [job_wrapper.create_job_parameter('inputXML', lua_xml,
                                                  cdata=True)]
    job_wrapper.run_job(vios_uuid, job_parms=job_parms)

    # Get the job result, and parse the output.
    job_result = job_wrapper.get_job_results_as_dict()
    status, devname, udid = _process_lua_result(job_result)
    return status, devname, udid


def _lua_recovery_xml(itls, adapter, vendor=LUAType.OTHER):
    """Builds the XML that is used as input for the lua_recovery job.

    The lua_recovery provides a very quick way for the system to discover
    an hdisk on the system.  This method builds the input into the lua_recovery
    job.
    :param itls: The list of ITL objects that define the various connections
                 between the server port (initiator), disk port (target) and
                 disk itself.
    :param vendor: The LUA vendor.  See the LUAType.* Constants.
    :return: The CDATA XML that is used for the lua_recovery job.
    """
    # Used for building the internal XML.

    root = ent.Element("XML_LIST", adapter, ns='')

    # The general attributes
    # TODO(IBM) Need to determine value of making these constants modifiable
    general = ent.Element("general", adapter, ns='')
    general.append(ent.Element("cmd_version", adapter, text=_LUA_CMD_VERSION,
                               ns=''))
    general.append(ent.Element("version", adapter, text=_LUA_VERSION, ns=''))
    root.append(general)

    # TODO(IBM) This can be re-evaluated.  Set to true if you know for sure
    # the ITLs are alive.  If there are any bad ITLs, this should be false.
    root.append(ent.Element("reliableITL", adapter, text="false", ns=''))

    # There is only one device in the device list.
    device_list = ent.Element("deviceList", adapter, ns='')
    device = ent.Element("device", adapter, ns='')
    device.append(ent.Element("vendor", adapter, text=vendor, ns=''))
    device.append(ent.Element("deviceTag", adapter, text="1", ns=''))

    itl_list = ent.Element("itlList", adapter, ns='')
    itl_list.append(ent.Element("number", adapter, text="%d" % (len(itls)),
                                ns=''))

    for itl in itls:
        itl_elem = ent.Element("itl", adapter, ns='')

        itl_elem.append(ent.Element("Iwwpn", adapter, text=itl.initiator,
                                    ns=''))
        itl_elem.append(ent.Element("Twwpn", adapter, text=itl.target, ns=''))
        itl_elem.append(ent.Element("lua", adapter, text=itl.lun, ns=''))

        itl_list.append(itl_elem)

    device.append(itl_list)
    device_list.append(device)
    root.append(device_list)

    return root.toxmlstring().decode('utf-8')


def _process_lua_result(result):
    """Processes the Output XML returned by LUARecovery.

    :return status: The status code from the discover process.
                    See LUAStatus.* constants.
    :return dev_name: The name of the discovered hdisk.
    :return udid: The UDID of the device.
    """
    if result is None:
        return None, None, None

    # The result may push to StdOut or to OutputXML (different versions push
    # to different locations).
    xml_resp = result.get('OutputXML')
    if xml_resp is None:
        xml_resp = result.get('StdOut')

    # If still none, nothing to do.
    if xml_resp is None:
        return None, None, None

    # The response is an XML block.  Put into an XML structure and get
    # the data out of it.
    root = etree.fromstring(xml_resp)
    base = 'deviceList/device/'
    estatus, edev_name, eudid, emessage = (
        root.find(base + x)
        for x in ('status', 'pvName', 'udid', 'msg/msgText'))
    status, dev_name, udid, message = (
        y.text if y is not None else None
        for y in (estatus, edev_name, eudid, emessage))
    _log_lua_status(status, dev_name, message)
    return status, dev_name, udid


def _log_lua_status(status, dev_name, message):
    """Logs any issues with the LUA."""

    if status == LUAStatus.DEVICE_AVAILABLE:
        LOG.info(_("LUA Recovery Successful. Device Found: %s"),
                 dev_name)
    elif status == LUAStatus.FOUND_ITL_ERR:
        # Message is already set.
        LOG.warning(_("ITL Error encountered: %s"), message)
    elif status == LUAStatus.DEVICE_IN_USE:
        LOG.warning(_("%s Device is currently in use."), dev_name)
    elif status == LUAStatus.FOUND_DEVICE_UNKNOWN_UDID:
        LOG.warning(_("%s Device discovered with unknown UDID."), dev_name)
    elif status == LUAStatus.INCORRECT_ITL:
        LOG.warning(_("Failed to Discover the Device : %s"), dev_name)


def remove_hdisk(adapter, host_name, dev_name, vios_uuid):
    """Command to remove the device from the VIOS.

    :param adapter: The pypowervm adapter.
    :param host_name: The name of the host.
    :param dev_name: The name of the device to remove.
    :param vios_uuid: The Virtual I/O Server UUID.
    """
    if adapter.traits.rmdev_job_available:
        _remove_hdisk_job(adapter, dev_name, vios_uuid)
    else:
        _remove_hdisk_classic(adapter, host_name, dev_name, vios_uuid)


def _remove_hdisk_job(adapter, dev_name, vios_uuid):
    """Runs the PowerVM Job to remove a hdisk.

    :param adapter: The pypowervm adapter.
    :param dev_name: The name of the device to remove.
    :param vios_uuid: The Virtual I/O Server UUID.
    """
    # Build up the job & invoke
    resp = adapter.read(
        pvm_vios.VIOS.schema_type, root_id=vios_uuid,
        suffix_type=c.SUFFIX_TYPE_DO, suffix_parm=_RM_HDISK)
    job_wrapper = pvm_job.Job.wrap(resp)
    job_parms = [job_wrapper.create_job_parameter('devName', dev_name)]

    # Run the job.  If the hdisk removal failed, the job will raise an
    # exception.  No output otherwise.
    job_wrapper.run_job(vios_uuid, job_parms=job_parms)


def _remove_hdisk_classic(adapter, host_name, dev_name, vios_uuid):
    """Command to remove the device from the VIOS.

    Runs a remote command to perform the action.

    :param adapter: The pypowervm adapter.
    :param host_name: The name of the host.
    :param dev_name: The name of the device to remove.
    :param vios_uuid: The Virtual I/O Server UUID.
    """
    try:
        # Execute a read on the vios to get the vios name
        resp = adapter.read(pvm_vios.VIOS.schema_type, root_id=vios_uuid)
        vios_w = pvm_vios.VIOS.wrap(resp)
        # build command
        rm_cmd = ('viosvrcmd -m ' + host_name + ' -p ' + vios_w.name +
                  ' -c \"rmdev -dev ' + dev_name + '\"')
        LOG.debug('RMDEV Command Input: %s' % rm_cmd)

        # Get the response for the CLIRunner command
        resp = adapter.read(_MGT_CONSOLE, None,
                            suffix_type=c.SUFFIX_TYPE_DO,
                            suffix_parm='CLIRunner')

        # Create the job parameters
        job_wrapper = pvm_job.Job.wrap(resp)
        ack_parm = 'acknowledgeThisAPIMayGoAwayInTheFuture'
        job_parms = [job_wrapper.create_job_parameter('cmd', rm_cmd),
                     job_wrapper.create_job_parameter(ack_parm,
                                                      'true')]

        job_wrapper.run_job(None, job_parms=job_parms)
        return job_wrapper.job_status()
    except pexc.JobRequestFailed as error:
        LOG.warning(_('CLIRunner Error: %s') % error)


def get_pg83_via_job(adapter, vios_uuid, udid):
    """Inventory call to fetch the encoded SCSI Page 0x83 descriptor for a PV.

    :param adapter: The pypowervm adapter through which to run the Job.
    :param vios_uuid: The UUID of the Virtual I/O Server owning the PV.
    :param udid: The UDID of the PV to query.
    :return: SCSI PG83 NAA descriptor, base64-encoded.  May be None.
    """
    # TODO(efried): Remove this method once VIOS supports pg83 in Events
    # Build the hdisk inventory input XML
    lua_xml = ('<uom:VIO xmlns:uom="http://www.ibm.com/xmlns/systems/power/fir'
               'mware/uom/mc/2012_10/" version="1.21" xmlns=""><uom:Request ac'
               'tion_str="QUERY_INVENTORY"><uom:InventoryRequest inventoryType'
               '="base"><uom:VioTypeFilter type="PV"/><uom:VioUdidFilter udid='
               '"%s"/></uom:InventoryRequest></uom:Request></uom:VIO>' % udid)

    # Build up the job & invoke
    job_wrapper = pvm_job.Job.wrap(adapter.read(
        pvm_vios.VIOS.schema_type, root_id=vios_uuid,
        suffix_type=c.SUFFIX_TYPE_DO, suffix_parm=_LUA_RECOVERY))
    job_wrapper.run_job(vios_uuid, job_parms=[
        job_wrapper.create_job_parameter('inputXML', lua_xml, cdata=True)])

    # Get the job result, and parse the output.
    result = job_wrapper.get_job_results_as_dict()

    # The result may push to StdOut or to OutputXML (different versions push
    # to different locations).
    if not result or not any((k in result for k in ('OutputXML', 'StdOut'))):
        LOG.warning(_('QUERY_INVENTORY LUARecovery Job succeeded, but result '
                      'contained neither OutputXML nor StdOut.'))
        return None
    xml_resp = result.get('OutputXML', result.get('StdOut'))
    LOG.debug('QUERY_INVENTORY result: %s' % xml_resp)

    return _parse_pg83_xml(xml_resp)


def _parse_pg83_xml(xml_resp):
    """Parse LUARecovery XML response, looking for pg83 descriptor.

    :param xml_resp: Tuple containing OutputXML and StdOut results of the
                     LUARecovery Job
    :return: pg83 descriptor text, or None if not found.
    """
    # QUERY_INVENTORY response may contain more than one element.  Each will be
    # delimited by its own <?xml?> tag.  etree will only parse one at a time.
    for chunk in xml_resp.split('<?xml version="1.0"?>'):
        if not chunk:
            continue
        try:
            parsed = etree.fromstring(chunk)
        except etree.XMLSyntaxError as e:
            LOG.warning(_('QUERY_INVENTORY produced invalid chunk of XML '
                          '(%(chunk)s).  Error: %(err)s'),
                        {'chunk': chunk, 'err': e.args[0]})
            continue
        for elem in parsed.getiterator():
            if (etree.QName(elem.tag).localname == 'PhysicalVolume_base' and
                    elem.attrib.get('desType') == "NAA"):
                return elem.attrib.get('descriptor')
    LOG.warning(_('Failed to find pg83 descriptor in XML output:\n%s'),
                xml_resp)
    return None
